from __future__ import annotations

import os

import pytest

from pbi_agent.agent.session import NEW_CHAT_SENTINEL, run_chat_loop, run_single_turn
from pbi_agent.config import DEFAULT_MODEL, Settings
from pbi_agent.models.messages import (
    CompletedResponse,
    TokenUsage,
    ToolCall,
    UserTurnInput,
)
from pbi_agent.session_store import SessionStore


class _DisplaySpy:
    def __init__(self) -> None:
        self.welcome_calls: list[dict[str, object | None]] = []
        self.session_usage_calls: list[TokenUsage] = []
        self.turn_usage_calls: list[tuple[TokenUsage, float]] = []
        self.debug_messages: list[str] = []
        self.reset_chat_calls = 0

    def welcome(
        self,
        *,
        interactive: bool = True,
        model: str | None = None,
        reasoning_effort: str | None = None,
        single_turn_hint: str | None = None,
    ) -> None:
        self.welcome_calls.append(
            {
                "interactive": interactive,
                "model": model,
                "reasoning_effort": reasoning_effort,
                "single_turn_hint": single_turn_hint,
            }
        )

    def session_usage(self, usage: TokenUsage) -> None:
        self.session_usage_calls.append(usage.snapshot())

    def turn_usage(self, usage: TokenUsage, elapsed_seconds: float) -> None:
        self.turn_usage_calls.append((usage.snapshot(), elapsed_seconds))

    def debug(self, message: str) -> None:
        self.debug_messages.append(message)

    def reset_chat(self) -> None:
        self.reset_chat_calls += 1

    def replay_history(self, messages) -> None:
        pass

    def begin_sub_agent(
        self,
        *,
        task_instruction: str,
        reasoning_effort: str | None = None,
        name: str = "sub_agent",
    ) -> _DisplaySpy:
        del task_instruction, reasoning_effort, name
        return self

    def finish_sub_agent(self, *, status: str) -> None:
        del status


class _ProviderStub:
    def __init__(self) -> None:
        self.connected = False
        self.request_calls: list[dict[str, object | None]] = []
        self.execute_calls: list[dict[str, object]] = []

    def __enter__(self) -> _ProviderStub:
        self.connect()
        return self

    def __exit__(self, *_: object) -> None:
        self.close()

    def connect(self) -> None:
        self.connected = True

    def close(self) -> None:
        self.connected = False

    def request_turn(
        self,
        *,
        user_message: str | None = None,
        user_input: UserTurnInput | None = None,
        tool_result_items: list[dict[str, object]] | None = None,
        instructions: str | None = None,
        display,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
    ) -> CompletedResponse:
        del display, instructions
        if user_input is not None and user_message is None:
            user_message = user_input.text
        self.request_calls.append(
            {
                "user_message": user_message,
                "tool_result_items": tool_result_items,
            }
        )
        if user_message is not None:
            response = CompletedResponse(
                response_id="resp_1",
                text="",
                usage=TokenUsage(input_tokens=5, output_tokens=1, model=DEFAULT_MODEL),
                function_calls=[
                    ToolCall(
                        call_id="call_1",
                        name="shell",
                        arguments={"command": "pwd"},
                    )
                ],
            )
        else:
            response = CompletedResponse(
                response_id="resp_2",
                text="All set.",
                usage=TokenUsage(input_tokens=3, output_tokens=2, model=DEFAULT_MODEL),
            )

        session_usage.add(response.usage)
        turn_usage.add(response.usage)
        return response

    def execute_tool_calls(
        self,
        response: CompletedResponse,
        *,
        max_workers: int,
        display,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
        sub_agent_depth: int = 0,
    ) -> tuple[list[dict[str, object]], bool]:
        del display, session_usage, turn_usage, sub_agent_depth
        self.execute_calls.append(
            {
                "response_id": response.response_id,
                "max_workers": max_workers,
                "call_count": len(response.function_calls),
            }
        )
        return (
            [
                {
                    "type": "function_call_output",
                    "call_id": "call_1",
                    "output": '{"ok": true, "result": "/workspace"}',
                }
            ],
            True,
        )


def test_run_single_turn_executes_tool_loop_and_aggregates_usage(monkeypatch) -> None:
    provider = _ProviderStub()
    display = _DisplaySpy()
    settings = Settings(api_key="test-key", provider="openai", max_tool_workers=3)
    monotonic_values = iter([10.0, 13.5])

    monkeypatch.setattr(
        "pbi_agent.agent.session.create_provider",
        lambda runtime_settings: provider,
    )
    monkeypatch.setattr(
        "pbi_agent.agent.session.time.monotonic",
        lambda: next(monotonic_values),
    )

    outcome = run_single_turn(
        "Inspect the workspace",
        settings,
        display,
        single_turn_hint="Single-turn test",
    )

    assert outcome.response_id == "resp_2"
    assert outcome.text == "All set."
    assert outcome.tool_errors is True
    assert provider.request_calls == [
        {
            "user_message": "Inspect the workspace",
            "tool_result_items": None,
        },
        {
            "user_message": None,
            "tool_result_items": [
                {
                    "type": "function_call_output",
                    "call_id": "call_1",
                    "output": '{"ok": true, "result": "/workspace"}',
                }
            ],
        },
    ]
    assert provider.execute_calls == [
        {"response_id": "resp_1", "max_workers": 3, "call_count": 1}
    ]
    assert display.welcome_calls == [
        {
            "interactive": False,
            "model": DEFAULT_MODEL,
            "reasoning_effort": "xhigh",
            "single_turn_hint": "Single-turn test",
        }
    ]
    assert display.debug_messages == ["model requested tool execution"]
    assert display.session_usage_calls[0].total_tokens == 0
    assert display.session_usage_calls[-1].input_tokens == 8
    assert display.session_usage_calls[-1].output_tokens == 3
    assert display.turn_usage_calls == [
        (
            TokenUsage(
                input_tokens=8,
                output_tokens=3,
                model=DEFAULT_MODEL,
            ),
            3.5,
        )
    ]


class _ChatDisplaySpy(_DisplaySpy):
    def __init__(self, prompts: list[str]) -> None:
        super().__init__()
        self.prompts = prompts
        self.prompt_calls = 0
        self.assistant_start_calls = 0

    def user_prompt(self) -> str:
        value = self.prompts[self.prompt_calls]
        self.prompt_calls += 1
        return value

    def assistant_start(self) -> None:
        self.assistant_start_calls += 1


class _ChatProviderStub:
    def __init__(self) -> None:
        self.request_messages: list[str | None] = []
        self.reset_calls = 0

    def __enter__(self) -> _ChatProviderStub:
        return self

    def __exit__(self, *_: object) -> None:
        return None

    def reset_conversation(self) -> None:
        self.reset_calls += 1

    def request_turn(
        self,
        *,
        user_message: str | None = None,
        user_input: UserTurnInput | None = None,
        tool_result_items: list[dict[str, object]] | None = None,
        instructions: str | None = None,
        display,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
    ) -> CompletedResponse:
        del display, instructions, tool_result_items
        if user_input is not None and user_message is None:
            user_message = user_input.text
        self.request_messages.append(user_message)
        response = CompletedResponse(
            response_id="resp_chat",
            text="Ack",
            usage=TokenUsage(input_tokens=2, output_tokens=1, model=DEFAULT_MODEL),
        )
        session_usage.add(response.usage)
        turn_usage.add(response.usage)
        return response


def test_run_chat_loop_resets_welcome_and_usage_on_new_chat(monkeypatch) -> None:
    provider = _ChatProviderStub()
    display = _ChatDisplaySpy(["hello", NEW_CHAT_SENTINEL, "after reset", "quit"])
    settings = Settings(api_key="test-key", provider="openai", max_tool_workers=2)
    monotonic_values = iter([5.0, 6.5, 10.0, 11.0])

    monkeypatch.setattr(
        "pbi_agent.agent.session.create_provider",
        lambda runtime_settings: provider,
    )
    monkeypatch.setattr(
        "pbi_agent.agent.session.time.monotonic",
        lambda: next(monotonic_values),
    )

    exit_code = run_chat_loop(settings, display)

    assert exit_code == 0
    assert provider.request_messages == ["hello", "after reset"]
    assert provider.reset_calls == 1
    assert len(display.welcome_calls) == 2
    assert display.welcome_calls[0]["interactive"] is True
    assert display.welcome_calls[1]["interactive"] is True
    assert [usage.total_tokens for usage in display.session_usage_calls] == [0, 3, 0, 3]
    assert display.reset_chat_calls == 1
    assert display.assistant_start_calls == 2


def test_run_chat_loop_does_not_persist_unanswered_user_turn(monkeypatch) -> None:
    class _FailingProvider:
        def __enter__(self):
            return self

        def __exit__(self, *_args):
            return None

        def reset_conversation(self) -> None:
            return None

        def request_turn(
            self,
            *,
            user_message: str | None = None,
            user_input: UserTurnInput | None = None,
            tool_result_items: list[dict[str, object]] | None = None,
            instructions: str | None = None,
            display,
            session_usage: TokenUsage,
            turn_usage: TokenUsage,
        ) -> CompletedResponse:
            del (
                user_message,
                user_input,
                tool_result_items,
                instructions,
                display,
                session_usage,
                turn_usage,
            )
            raise RuntimeError("boom")

    display = _ChatDisplaySpy(["hello"])
    settings = Settings(api_key="test-key", provider="openai", max_tool_workers=2)

    monkeypatch.setattr(
        "pbi_agent.agent.session.create_provider",
        lambda runtime_settings: _FailingProvider(),
    )

    with pytest.raises(RuntimeError, match="boom"):
        run_chat_loop(settings, display)

    with SessionStore() as store:
        sessions = store.list_sessions(os.getcwd(), limit=10)
        assert len(sessions) == 1
        assert store.list_messages(sessions[0].session_id) == []



from __future__ import annotations

import json
import urllib.error

from email.message import Message
from io import BytesIO
from typing import Any, Callable

import pytest

from pbi_agent.models.messages import TokenUsage, WebSearchSource


class DisplaySpy:
    def __init__(self) -> None:
        self.wait_messages: list[str] = []
        self.wait_stop_calls = 0
        self.retry_notices: list[tuple[int, int]] = []
        self.rate_limit_notices: list[tuple[float, int, int]] = []
        self.overload_notices: list[tuple[float, int, int]] = []
        self.session_usage_snapshots: list[TokenUsage] = []
        self.thinking_calls: list[dict[str, object | None]] = []
        self.redacted_thinking_calls = 0
        self.markdown_calls: list[str] = []
        self.function_counts: list[int] = []
        self.function_results: list[dict[str, object]] = []
        self.tool_group_end_count = 0
        self.web_search_sources_calls: list[list[WebSearchSource]] = []

    def begin_sub_agent(
        self,
        *,
        task_instruction: str,
        reasoning_effort: str | None = None,
        name: str = "sub_agent",
    ) -> DisplaySpy:
        del task_instruction, reasoning_effort, name
        return self

    def finish_sub_agent(self, *, status: str) -> None:
        del status

    def wait_start(self, message: str = "") -> None:
        self.wait_messages.append(message)

    def wait_stop(self) -> None:
        self.wait_stop_calls += 1

    def retry_notice(self, attempt: int, max_retries: int) -> None:
        self.retry_notices.append((attempt, max_retries))

    def rate_limit_notice(
        self,
        *,
        wait_seconds: float,
        attempt: int,
        max_retries: int,
    ) -> None:
        self.rate_limit_notices.append((wait_seconds, attempt, max_retries))

    def overload_notice(
        self,
        *,
        wait_seconds: float,
        attempt: int,
        max_retries: int,
    ) -> None:
        self.overload_notices.append((wait_seconds, attempt, max_retries))

    def session_usage(self, usage: TokenUsage) -> None:
        self.session_usage_snapshots.append(usage.snapshot())

    def render_thinking(
        self,
        text: str | None = None,
        *,
        title: str | None = None,
        replace_existing: bool = False,
        widget_id: str | None = None,
    ) -> str | None:
        self.thinking_calls.append(
            {
                "text": text,
                "title": title,
                "replace_existing": replace_existing,
                "widget_id": widget_id,
            }
        )
        return widget_id or f"thinking-{len(self.thinking_calls)}"

    def render_redacted_thinking(self) -> None:
        self.redacted_thinking_calls += 1

    def render_markdown(self, text: str) -> None:
        self.markdown_calls.append(text)

    def function_start(self, count: int) -> None:
        self.function_counts.append(count)

    def function_result(
        self,
        *,
        name: str,
        success: bool,
        call_id: str,
        arguments: object,
    ) -> None:
        self.function_results.append(
            {
                "name": name,
                "success": success,
                "call_id": call_id,
                "arguments": arguments,
            }
        )

    def tool_group_end(self) -> None:
        self.tool_group_end_count += 1

    def web_search_sources(self, sources: list[WebSearchSource]) -> None:
        self.web_search_sources_calls.append(list(sources))


class FakeHTTPResponse:
    def __init__(self, payload: dict[str, Any]) -> None:
        self._payload = payload

    def read(self) -> bytes:
        return json.dumps(self._payload).encode("utf-8")

    def __enter__(self) -> FakeHTTPResponse:
        return self

    def __exit__(self, *_: object) -> None:
        return None


@pytest.fixture
def display_spy() -> DisplaySpy:
    return DisplaySpy()


@pytest.fixture(autouse=True)
def isolate_internal_config_path(monkeypatch: pytest.MonkeyPatch, tmp_path) -> None:
    monkeypatch.setenv(
        "PBI_AGENT_INTERNAL_CONFIG_PATH", str(tmp_path / "internal-config.json")
    )


@pytest.fixture(autouse=True)
def isolate_session_db(monkeypatch: pytest.MonkeyPatch, tmp_path) -> None:
    monkeypatch.setenv("PBI_AGENT_SESSION_DB_PATH", str(tmp_path / "sessions.db"))


@pytest.fixture
def make_http_response() -> Callable[[dict[str, Any]], FakeHTTPResponse]:
    def factory(payload: dict[str, Any]) -> FakeHTTPResponse:
        return FakeHTTPResponse(payload)

    return factory


@pytest.fixture
def make_http_error() -> Callable[..., urllib.error.HTTPError]:
    def factory(
        *,
        url: str = "https://example.invalid",
        code: int,
        body: str = "",
        headers: dict[str, str] | None = None,
    ) -> urllib.error.HTTPError:
        message = Message()
        for key, value in (headers or {}).items():
            message[key] = str(value)
        return urllib.error.HTTPError(
            url=url,
            code=code,
            msg=f"HTTP {code}",
            hdrs=message,
            fp=BytesIO(body.encode("utf-8")),
        )

    return factory



"""Tests for Display tool-call formatting methods."""

from __future__ import annotations

import threading
from unittest.mock import MagicMock

from pbi_agent.agent.session import NEW_CHAT_SENTINEL
from pbi_agent.models.messages import TokenUsage
from pbi_agent.ui.display import Display
from pbi_agent.ui.formatting import (
    format_generic_function_item,
    format_init_report_item,
    format_skill_knowledge_item,
    format_web_search_sources_item,
)


def _make_display(*, verbose: bool = False) -> Display:
    """Create a Display with a stub app for unit-testing format helpers."""
    app = MagicMock()
    return Display(app, verbose=verbose)


# -- format_skill_knowledge_item --------------------------------------------


class TestFormatSkillKnowledgeItem:
    def test_single_skill(self) -> None:
        result = format_skill_knowledge_item(
            ["card_visual"], status="[green]done[/green]"
        )
        assert "card_visual" in result
        assert "[green]done[/green]" in result

    def test_multiple_skills(self) -> None:
        result = format_skill_knowledge_item(
            ["card_visual", "table_visual"], status="[green]done[/green]"
        )
        assert "card_visual, table_visual" in result

    def test_empty_skills_shows_none(self) -> None:
        result = format_skill_knowledge_item([], status="[green]done[/green]")
        assert "<none>" in result

    def test_verbose_includes_call_id(self) -> None:
        result = format_skill_knowledge_item(
            ["card_visual"],
            verbose=True,
            status="[green]done[/green]",
            call_id="call_42",
        )
        assert "call_42" in result

    def test_non_verbose_omits_call_id(self) -> None:
        result = format_skill_knowledge_item(
            ["card_visual"],
            verbose=False,
            status="[green]done[/green]",
            call_id="call_42",
        )
        assert "call_42" not in result


# -- format_init_report_item ------------------------------------------------


class TestFormatInitReportItem:
    def test_shows_destination(self) -> None:
        result = format_init_report_item(".", status="[green]done[/green]")
        assert "." in result
        assert "[green]done[/green]" in result

    def test_force_true_shown(self) -> None:
        result = format_init_report_item(
            "/tmp/report", status="[green]done[/green]", force=True
        )
        assert "force" in result
        assert "true" in result

    def test_force_false_omitted(self) -> None:
        result = format_init_report_item(
            "/tmp/report", status="[green]done[/green]", force=False
        )
        assert "force" not in result

    def test_verbose_includes_call_id(self) -> None:
        result = format_init_report_item(
            ".", verbose=True, status="[green]done[/green]", call_id="call_7"
        )
        assert "call_7" in result


# -- format_generic_function_item -------------------------------------------


class TestFormatGenericFunctionItem:
    def test_non_verbose_with_args_shows_summary(self) -> None:
        result = format_generic_function_item(
            "my_tool",
            verbose=False,
            status="[green]done[/green]",
            arguments={"key": "value"},
        )
        assert "my_tool()" in result
        assert "key" in result
        assert "value" in result

    def test_non_verbose_no_args_shows_name_only(self) -> None:
        result = format_generic_function_item(
            "my_tool",
            verbose=False,
            status="[green]done[/green]",
            arguments=None,
        )
        assert "my_tool()" in result
        assert "[green]done[/green]" in result
        assert "\n" not in result

    def test_non_verbose_empty_dict_args_shows_name_only(self) -> None:
        result = format_generic_function_item(
            "my_tool",
            verbose=False,
            status="[green]done[/green]",
            arguments={},
        )
        assert "my_tool()" in result
        assert "[green]done[/green]" in result
        assert "\n" not in result

    def test_verbose_includes_call_id_and_args(self) -> None:
        result = format_generic_function_item(
            "my_tool",
            verbose=True,
            status="[green]done[/green]",
            call_id="call_99",
            arguments={"key": "value"},
        )
        assert "call_id=call_99" in result
        assert "args=" in result


class TestFormatWebSearchSourcesItem:
    def test_includes_queries_when_present(self) -> None:
        result = format_web_search_sources_item(
            [{"title": "BTC", "url": "https://example.com/btc", "snippet": ""}],
            queries=["finance: BTC"],
            status="[green]done[/green]",
        )
        assert "queries:" in result
        assert "finance: BTC" in result
        assert "web search" not in result
        assert "🔍" not in result

    def test_no_sources_still_shows_queries(self) -> None:
        result = format_web_search_sources_item(
            [],
            queries=["finance: BTC"],
            status="[green]done[/green]",
        )
        assert "no sources" in result
        assert "finance: BTC" in result

    def test_queries_render_before_source_summary(self) -> None:
        result = format_web_search_sources_item(
            [{"title": "BTC", "url": "https://example.com/btc", "snippet": ""}],
            queries=["finance: BTC"],
            status="[green]done[/green]",
        )
        assert result.index("queries:") < result.index("1 source")

    def test_compact_sources_use_separate_url_line(self) -> None:
        result = format_web_search_sources_item(
            [
                {
                    "title": "CoinMarketCap",
                    "url": "https://example.com/btc",
                    "snippet": "",
                }
            ],
            status="[green]done[/green]",
        )
        assert "[dim]•[/dim] CoinMarketCap" in result
        assert "https://example.com/btc" in result
        assert "CoinMarketCap  [dim]https://example.com/btc[/dim]" not in result


class TestRenderThinking:
    def test_missing_body_uses_summary_text(self) -> None:
        app = MagicMock()
        display = Display(app)
        summary = (
            "I need to summarize the content, possibly mentioning if there are any "
            "continuities present. This full text should reach the widget body."
        )

        display.render_thinking(title=summary)

        assert app.call_from_thread.call_count == 1
        _, widget_id, widget_title, widget_body = app.call_from_thread.call_args.args
        assert widget_id == "thinking-1"
        assert widget_title == "Thinking..."
        assert widget_body == summary

    def test_ellipsis_body_uses_summary_text(self) -> None:
        app = MagicMock()
        display = Display(app)
        summary = (
            "I need to summarize the content, possibly mentioning if there are any "
            "continuities present. This should replace ellipsis-only content."
        )

        display.render_thinking("...", title=summary)

        assert app.call_from_thread.call_count == 1
        _, _, widget_title, widget_body = app.call_from_thread.call_args.args
        assert widget_title == "Thinking..."
        assert widget_body == summary


class TestInteractiveInputQueue:
    def test_new_chat_request_queued_before_prompt_is_preserved(self) -> None:
        display = _make_display()
        result: list[str] = []

        display.request_new_chat()

        worker = threading.Thread(target=lambda: result.append(display.user_prompt()))
        worker.start()
        worker.join(timeout=1)

        assert not worker.is_alive()
        assert result == [NEW_CHAT_SENTINEL]


def test_session_usage_updates_header_with_model_and_effort() -> None:
    app = MagicMock()
    display = Display(app, model="gpt-5.4-2026-03-05", reasoning_effort="xhigh")

    display.session_usage(
        TokenUsage(input_tokens=8, output_tokens=3, model="gpt-5.4-2026-03-05")
    )

    callback, subtitle = app.call_from_thread.call_args.args
    assert callback == app.update_session_header
    assert "gpt-5.4-2026-03-05 (xhigh)" in subtitle
    assert "11 tok" in subtitle
    assert "ctx" not in subtitle
    assert app.call_from_thread.call_args.kwargs["context_label"] is None


# -- function_result routing -------------------------------------------------


class TestFunctionResultRouting:
    """Verify function_result routes skill_knowledge and init_report properly."""

    def test_skill_knowledge_routed(self) -> None:
        d = _make_display()
        d.function_result(
            "skill_knowledge",
            True,
            call_id="call_sk",
            arguments={"skills": ["card_visual", "table_visual"]},
        )
        assert len(d._tool_group.items) == 1
        text = d._tool_group.items[0].text
        assert "card_visual" in text
        assert "table_visual" in text

    def test_init_report_routed(self) -> None:
        d = _make_display()
        d.function_result(
            "init_report",
            True,
            call_id="call_ir",
            arguments={"dest": "/my/report", "force": True},
        )
        assert len(d._tool_group.items) == 1
        text = d._tool_group.items[0].text
        assert "/my/report" in text
        assert "force" in text

    def test_generic_function_routed(self) -> None:
        d = _make_display()
        d.function_result(
            "some_unknown_tool",
            False,
            call_id="call_u",
            arguments={"x": 1},
        )
        assert len(d._tool_group.items) == 1
        text = d._tool_group.items[0].text
        assert "some_unknown_tool()" in text

    def test_shell_still_routed(self) -> None:
        d = _make_display()
        d.function_result(
            "shell",
            True,
            call_id="call_sh",
            arguments={"command": "ls -la"},
        )
        assert len(d._tool_group.items) == 1
        text = d._tool_group.items[0].text
        assert "ls -la" in text

    def test_apply_patch_still_routed(self) -> None:
        d = _make_display()
        d.function_result(
            "apply_patch",
            True,
            call_id="call_ap",
            arguments={
                "path": "/tmp/file.txt",
                "operation_type": "create",
                "diff": "--- a\n+++ b",
            },
        )
        assert len(d._tool_group.items) == 1
        text = d._tool_group.items[0].text
        assert "/tmp/file.txt" in text
        assert "create" in text


def test_begin_sub_agent_mounts_nested_block_and_child_widgets() -> None:
    app = MagicMock()
    display = Display(app)

    sub_display = display.begin_sub_agent(
        task_instruction="Inspect source files for TODOs",
        reasoning_effort="low",
        name="Hermes",
    )
    sub_display.render_markdown("Found one TODO.")
    sub_display.finish_sub_agent(status="completed")

    first_call = app.call_from_thread.call_args_list[0]
    assert first_call.args[0] == app.mount_sub_agent_block
    assert "Hermes" in first_call.args[2]
    assert "Inspect source files for TODOs" in first_call.args[2]

    callbacks = [call.args[0] for call in app.call_from_thread.call_args_list]
    assert app.mount_widget_in_container in callbacks
    assert app.update_sub_agent_title in callbacks


def test_sub_agent_render_thinking_queries_via_call_from_thread() -> None:
    app = MagicMock()

    def call_from_thread(callback, *args, **kwargs):  # type: ignore[no-untyped-def]
        if callback is app._query_optional:
            return None
        return None

    app.call_from_thread.side_effect = call_from_thread
    display = Display(app)

    sub_display = display.begin_sub_agent(
        task_instruction="Inspect source files for TODOs",
        reasoning_effort="low",
        name="Hermes",
    )
    sub_display.render_thinking("Reasoning details", title="Thinking")

    callbacks = [call.args[0] for call in app.call_from_thread.call_args_list]
    assert app._query_optional in callbacks
    assert app.mount_widget_in_container in callbacks



from __future__ import annotations

from pathlib import Path
from typing import Any

from pbi_agent.media import load_workspace_image
from pbi_agent.tools.types import ToolContext, ToolOutput, ToolSpec

SPEC = ToolSpec(
    name="read_image",
    description=(
        "Read a workspace image file safely and attach it to the model context in "
        "native multimodal format, alongside a compact metadata summary."
    ),
    parameters_schema={
        "type": "object",
        "properties": {
            "path": {
                "type": "string",
                "description": (
                    "Image path relative to the workspace root "
                    "(or absolute within workspace)."
                ),
            }
        },
        "required": ["path"],
        "additionalProperties": False,
    },
)


def handle(
    arguments: dict[str, Any], context: ToolContext
) -> dict[str, Any] | ToolOutput:
    del context
    path_value = arguments.get("path", "")
    if not isinstance(path_value, str) or not path_value.strip():
        return {"error": "'path' must be a non-empty string."}

    root = Path.cwd().resolve()
    try:
        image = load_workspace_image(root, path_value)
    except Exception as exc:
        return {"error": str(exc)}

    summary = {
        "path": image.path,
        "mime_type": image.mime_type,
        "byte_count": image.byte_count,
    }
    return ToolOutput(result=summary, attachments=[image])



from __future__ import annotations

import pytest

from pbi_agent.config import Settings
from pbi_agent.providers import create_provider
from pbi_agent.providers.anthropic_provider import AnthropicProvider
from pbi_agent.providers.generic_provider import GenericProvider
from pbi_agent.providers.google_provider import GoogleProvider
from pbi_agent.providers.openai_provider import OpenAIProvider
from pbi_agent.providers.xai_provider import XAIProvider


@pytest.mark.parametrize(
    ("provider_name", "expected_type"),
    [
        ("openai", OpenAIProvider),
        ("xai", XAIProvider),
        ("google", GoogleProvider),
        ("anthropic", AnthropicProvider),
        ("generic", GenericProvider),
    ],
)
def test_create_provider_returns_expected_backend(
    provider_name: str,
    expected_type: type,
) -> None:
    settings = Settings(api_key="test-key", provider=provider_name)

    provider = create_provider(settings)

    assert isinstance(provider, expected_type)



from __future__ import annotations

import json
import re
from dataclasses import dataclass
from typing import Any

_GENERIC_ERROR_MESSAGES = {
    "provider returned error",
    "no error message",
}
_UNRECOGNIZED_CHAT_MESSAGE = "Unrecognized chat message."
_MODEL_UNAVAILABLE_MESSAGE = (
    "No allowed providers are available for the selected model."
)
_WHITESPACE_RE = re.compile(r"\s+")


@dataclass(slots=True)
class _StructuredErrorInfo:
    error_type: str | None = None
    message: str = ""
    request_id: str | None = None


def format_user_facing_error(exc: BaseException) -> str:
    raw_message = _normalize_whitespace(str(exc).strip())
    if not raw_message:
        return f"{exc.__class__.__name__}."

    structured = _extract_structured_error(raw_message)
    detail = _normalize_whitespace(structured.message)
    error_type = (structured.error_type or "").lower()
    request_id = structured.request_id

    if error_type == "overloaded_error" or "overloaded" in raw_message.lower():
        parts = ["Provider overloaded."]
        _append_detail(
            parts,
            detail,
            ignored={"overloaded", "the api is temporarily overloaded."},
        )
        parts.append("Retry shortly.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "resource_exhausted":
        parts = ["Rate limit reached."]
        _append_detail(parts, detail)
        parts.append("Check Gemini API rate limits or request a quota increase.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "rate_limit_error" or "rate limit" in raw_message.lower():
        parts = ["Rate limit reached."]
        _append_detail(parts, detail)
        parts.append("Retry shortly or switch model/provider.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "failed_precondition":
        parts = ["Request cannot be served in the current project or region."]
        _append_detail(parts, detail)
        parts.append(
            "If Gemini free tier is unavailable in your region, enable billing in Google AI Studio."
        )
        return _finalize_message(parts, request_id=request_id)

    if _contains_phrase(raw_message, detail, _UNRECOGNIZED_CHAT_MESSAGE):
        return _finalize_message(
            [
                "Provider rejected the tool follow-up request.",
                _preferred_detail(raw_message, detail, _UNRECOGNIZED_CHAT_MESSAGE),
                "This model/provider combination may not support the current tool-call transcript.",
            ],
            request_id=request_id,
        )

    if _contains_phrase(raw_message, detail, _MODEL_UNAVAILABLE_MESSAGE):
        return _finalize_message(
            [
                "Selected model is not available through the configured provider.",
                _preferred_detail(raw_message, detail, _MODEL_UNAVAILABLE_MESSAGE),
                "Choose a model/provider combination that is available on your gateway.",
            ],
            request_id=request_id,
        )

    if error_type == "authentication_error":
        parts = ["Authentication failed."]
        _append_detail(
            parts,
            detail,
            ignored={"there's an issue with your api key."},
        )
        parts.append("Check the configured API key.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "permission_error":
        parts = ["Permission denied."]
        _append_detail(
            parts,
            detail,
            ignored={
                "your api key does not have permission to use the specified resource."
            },
        )
        parts.append("Check that the API key can access the requested resource.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "permission_denied":
        parts = ["Permission denied."]
        _append_detail(parts, detail)
        parts.append("Check the Gemini API key permissions and model access.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "not_found_error":
        parts = ["Requested resource was not found."]
        _append_detail(
            parts,
            detail,
            ignored={"the requested resource could not be found."},
        )
        return _finalize_message(parts, request_id=request_id)

    if error_type == "not_found":
        parts = ["Requested resource was not found."]
        _append_detail(parts, detail)
        parts.append(
            "Check the request resource identifiers, model name, and API version."
        )
        return _finalize_message(parts, request_id=request_id)

    if error_type == "request_too_large":
        parts = ["Request is too large."]
        _append_detail(
            parts,
            detail,
            ignored={"request exceeds the maximum allowed number of bytes."},
        )
        parts.append("Reduce the Anthropic Messages API request size below 32 MB.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "invalid_request_error":
        parts = ["Invalid request."]
        _append_detail(
            parts,
            detail,
            ignored={"there was an issue with the format or content of your request."},
        )
        parts.append("Check the request payload, selected model, and tool transcript.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "invalid_argument":
        parts = ["Invalid request."]
        _append_detail(parts, detail)
        parts.append("Check Gemini request fields, model parameters, and API version.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "api_error":
        parts = ["Provider error."]
        _append_detail(
            parts,
            detail,
            ignored={
                "an unexpected error has occurred internal to anthropic's systems."
            },
        )
        parts.append("Retry shortly.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "internal":
        parts = ["Provider error."]
        _append_detail(parts, detail)
        parts.append(
            "Reduce the prompt/context size or retry with another Gemini model."
        )
        return _finalize_message(parts, request_id=request_id)

    if error_type == "unavailable":
        parts = ["Provider overloaded."]
        _append_detail(parts, detail)
        parts.append("Retry shortly or switch to another Gemini model.")
        return _finalize_message(parts, request_id=request_id)

    if error_type == "deadline_exceeded":
        parts = ["Request timed out."]
        _append_detail(parts, detail)
        parts.append("Increase the client timeout or reduce the prompt/context size.")
        return _finalize_message(parts, request_id=request_id)

    if "api key was reported as leaked" in raw_message.lower() or (
        detail and "api key was reported as leaked" in detail.lower()
    ):
        return _finalize_message(
            [
                "Authentication failed.",
                "Gemini rejected an API key that was reported as leaked.",
                "Generate a new API key in Google AI Studio and replace the blocked key.",
            ],
            request_id=request_id,
        )

    if detail and detail != raw_message:
        return _finalize_message(["Request failed.", detail], request_id=request_id)

    return _finalize_message([raw_message], request_id=request_id)


def _extract_structured_error(message: str) -> _StructuredErrorInfo:
    payload = _parse_embedded_json(message)
    if payload is None:
        return _StructuredErrorInfo(message=message)

    request_id = payload.get("request_id")
    error_type: str | None = None
    detail = ""

    error_value = payload.get("error")
    if isinstance(error_value, dict):
        payload_type = error_value.get("type")
        if isinstance(payload_type, str) and payload_type.strip():
            error_type = payload_type.strip()
        elif (
            isinstance(error_value.get("status"), str)
            and error_value.get("status").strip()
        ):
            error_type = error_value.get("status").strip().lower()
        payload_message = error_value.get("message")
        if isinstance(payload_message, str) and payload_message.strip():
            detail = payload_message.strip()
    elif isinstance(error_value, str) and error_value.strip():
        detail = error_value.strip()

    if not detail:
        detail = _extract_from_payload(payload) or message

    return _StructuredErrorInfo(
        error_type=error_type,
        message=detail,
        request_id=request_id.strip()
        if isinstance(request_id, str) and request_id.strip()
        else None,
    )


def _extract_error_detail(message: str) -> str:
    payload = _parse_embedded_json(message)
    if payload is None:
        return message

    extracted = _extract_from_payload(payload)
    return extracted or message


def _parse_embedded_json(message: str) -> dict[str, Any] | None:
    start = message.find("{")
    if start < 0:
        return None
    try:
        payload = json.loads(message[start:])
    except json.JSONDecodeError:
        return None
    return payload if isinstance(payload, dict) else None


def _extract_from_payload(payload: dict[str, Any]) -> str:
    candidates: list[str] = []
    _collect_candidate_messages(payload, candidates)

    seen: set[str] = set()
    filtered: list[str] = []
    for candidate in candidates:
        normalized = _normalize_whitespace(candidate)
        if not normalized:
            continue
        if normalized.lower() in _GENERIC_ERROR_MESSAGES:
            continue
        if normalized in seen:
            continue
        seen.add(normalized)
        filtered.append(normalized)

    return "\n".join(filtered[:2])


def _collect_candidate_messages(value: Any, candidates: list[str]) -> None:
    if isinstance(value, dict):
        message = value.get("message")
        if isinstance(message, str):
            candidates.append(message)

        metadata = value.get("metadata")
        if isinstance(metadata, dict):
            raw_value = metadata.get("raw")
            if isinstance(raw_value, str):
                parsed_raw = _parse_nested_json_string(raw_value)
                if parsed_raw is not None:
                    _collect_candidate_messages(parsed_raw, candidates)
                else:
                    candidates.append(raw_value)

            available = metadata.get("available_providers")
            if isinstance(available, list) and available:
                providers = ", ".join(str(item) for item in available if str(item))
                if providers:
                    candidates.append(f"Available providers: {providers}.")

        error_value = value.get("error")
        if isinstance(error_value, str):
            candidates.append(error_value)
        elif error_value is not None and error_value is not value:
            _collect_candidate_messages(error_value, candidates)
        return

    if isinstance(value, list):
        for item in value:
            _collect_candidate_messages(item, candidates)


def _parse_nested_json_string(raw_value: str) -> dict[str, Any] | None:
    stripped = raw_value.strip()
    if not stripped.startswith("{"):
        return None
    try:
        parsed = json.loads(stripped)
    except json.JSONDecodeError:
        return None
    return parsed if isinstance(parsed, dict) else None


def _normalize_whitespace(text: str) -> str:
    return _WHITESPACE_RE.sub(" ", text).strip()


def _contains_phrase(raw_message: str, detail: str, phrase: str) -> bool:
    lowered_phrase = phrase.lower()
    return lowered_phrase in raw_message.lower() or lowered_phrase in detail.lower()


def _preferred_detail(raw_message: str, detail: str, phrase: str) -> str:
    return phrase if phrase.lower() in raw_message.lower() else detail


def _append_detail(
    parts: list[str],
    detail: str,
    *,
    ignored: set[str] | None = None,
) -> None:
    normalized = _normalize_whitespace(detail)
    if not normalized:
        return
    if normalized == parts[0]:
        return
    if ignored and normalized.lower() in ignored:
        return
    parts.append(normalized)


def _finalize_message(parts: list[str], *, request_id: str | None) -> str:
    final_parts = [part for part in parts if part]
    if request_id:
        request_line = f"Request ID: {request_id}"
        if request_line not in final_parts:
            final_parts.append(request_line)
    return "\n".join(final_parts)



from __future__ import annotations

import asyncio
from io import StringIO
from unittest.mock import Mock

from rich.console import Console

from pbi_agent.branding import PBI_AGENT_NAME, PBI_AGENT_TAGLINE
from pbi_agent.web.serve import _FaviconServer


def test_favicon_server_startup_uses_pbi_agent_banner() -> None:
    server = _FaviconServer(command="uv run pbi-agent web")
    output = StringIO()
    server.console = Console(file=output, width=80, highlight=False)

    asyncio.run(server.on_startup(Mock()))

    rendered = output.getvalue()
    assert PBI_AGENT_NAME in rendered
    assert PBI_AGENT_TAGLINE in rendered
    assert "textual-serve" not in rendered
    assert server.public_url in rendered



from __future__ import annotations

from unittest.mock import ANY

import pytest

from pbi_agent.agent.session import run_sub_agent_task
from pbi_agent.models.messages import CompletedResponse, TokenUsage
from pbi_agent.tools import sub_agent as sub_agent_tool
from pbi_agent.tools.types import ToolContext
from pbi_agent.config import Settings


class _ChildDisplay:
    def __init__(self) -> None:
        self.finished_statuses: list[str] = []

    def begin_sub_agent(
        self,
        *,
        task_instruction: str,
        reasoning_effort: str | None = None,
        name: str = "sub_agent",
    ) -> _ChildDisplay:
        del task_instruction, reasoning_effort, name
        return self

    def finish_sub_agent(self, *, status: str) -> None:
        self.finished_statuses.append(status)

    def session_usage(self, usage: TokenUsage) -> None:
        self.session_usage_snapshot = usage.snapshot()

    def turn_usage(self, usage: TokenUsage, elapsed_seconds: float) -> None:
        self.turn_usage_snapshot = (usage.snapshot(), elapsed_seconds)

    def wait_start(self, message: str = "") -> None:
        self.last_wait_message = message

    def wait_stop(self) -> None:
        self.wait_stopped = True

    def render_markdown(self, text: str) -> None:
        self.markdown = text

    def render_thinking(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        del args, kwargs
        return None

    def render_redacted_thinking(self) -> None:
        self.redacted = True

    def function_start(self, count: int) -> None:
        self.function_count = count

    def function_result(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        del args, kwargs

    def tool_group_end(self) -> None:
        self.tool_group_closed = True

    def retry_notice(self, attempt: int, max_retries: int) -> None:
        self.retry = (attempt, max_retries)

    def rate_limit_notice(
        self, *, wait_seconds: float, attempt: int, max_retries: int
    ) -> None:
        self.rate_limit = (wait_seconds, attempt, max_retries)

    def overload_notice(
        self, *, wait_seconds: float, attempt: int, max_retries: int
    ) -> None:
        self.overload = (wait_seconds, attempt, max_retries)

    def error(self, message: str) -> None:
        self.error_message = message

    def debug(self, message: str) -> None:
        self.debug_message = message

    def request_shutdown(self) -> None:
        return None

    def submit_input(self, value: str) -> None:
        del value

    def request_new_chat(self) -> None:
        return None

    def reset_chat(self) -> None:
        return None

    def welcome(self, **kwargs) -> None:  # type: ignore[no-untyped-def]
        del kwargs

    def user_prompt(self) -> str:
        raise RuntimeError("not supported")

    def assistant_start(self) -> None:
        return None


class _ParentDisplay:
    def __init__(self) -> None:
        self.child_display = _ChildDisplay()
        self.sub_agent_calls: list[dict[str, str | None]] = []
        self.session_usage_calls: list[TokenUsage] = []

    def begin_sub_agent(
        self,
        *,
        task_instruction: str,
        reasoning_effort: str | None = None,
        name: str = "sub_agent",
    ) -> _ChildDisplay:
        self.sub_agent_calls.append(
            {
                "task_instruction": task_instruction,
                "reasoning_effort": reasoning_effort,
                "name": name,
            }
        )
        return self.child_display

    def finish_sub_agent(self, *, status: str) -> None:
        del status

    def session_usage(self, usage: TokenUsage) -> None:
        self.session_usage_calls.append(usage.snapshot())


class _ProviderStub:
    def __init__(self) -> None:
        self.request_messages: list[str | None] = []

    def __enter__(self) -> _ProviderStub:
        return self

    def __exit__(self, *_: object) -> None:
        return None

    def request_turn(
        self,
        *,
        user_message: str | None = None,
        tool_result_items=None,
        instructions: str | None = None,
        display,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
    ) -> CompletedResponse:
        del tool_result_items, display
        self.request_messages.append(user_message)
        self.instructions = instructions
        response = CompletedResponse(
            response_id="resp_child",
            text="child complete",
            usage=TokenUsage(input_tokens=4, output_tokens=2, model="gpt-5"),
        )
        session_usage.add(response.usage)
        turn_usage.add(response.usage)
        return response

    def execute_tool_calls(
        self,
        response: CompletedResponse,
        *,
        max_workers: int,
        display,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
        sub_agent_depth: int = 0,
    ):
        del response, max_workers, display, session_usage, turn_usage, sub_agent_depth
        return [], False


def test_sub_agent_tool_blocks_nested_calls() -> None:
    result = sub_agent_tool.handle(
        {"task_instruction": "Inspect src", "reasoning_effort": "medium"},
        ToolContext(sub_agent_depth=1),
    )

    assert result["status"] == "failed"
    assert result["error"]["type"] == "nested_sub_agent_disabled"


@pytest.mark.parametrize(
    ("sub_agent_model", "expected_model"),
    [
        ("gpt-5-mini", "gpt-5-mini"),
        (None, "gpt-5"),
    ],
)
def test_run_sub_agent_task_uses_child_prompt_and_aggregates_usage(
    monkeypatch, sub_agent_model: str | None, expected_model: str
) -> None:
    captured: dict[str, object] = {}

    def fake_create_provider(
        settings: Settings,
        *,
        system_prompt: str | None = None,
        excluded_tools: set[str] | None = None,
    ) -> _ProviderStub:
        captured["settings"] = settings
        captured["system_prompt"] = system_prompt
        captured["excluded_tools"] = excluded_tools
        return _ProviderStub()

    monkeypatch.setattr("pbi_agent.agent.session.create_provider", fake_create_provider)

    parent_display = _ParentDisplay()
    parent_session_usage = TokenUsage(model="gpt-5")
    parent_turn_usage = TokenUsage(model="gpt-5")
    settings = Settings(
        api_key="test-key",
        provider="openai",
        model="gpt-5",
        sub_agent_model=sub_agent_model,
    )

    result = run_sub_agent_task(
        "Summarize the repo structure",
        settings,
        parent_display,
        reasoning_effort="medium",
        parent_session_usage=parent_session_usage,
        parent_turn_usage=parent_turn_usage,
    )

    assert result == {
        "status": "completed",
        "final_output": "child complete",
    }
    assert parent_display.sub_agent_calls == [
        {
            "task_instruction": "Summarize the repo structure",
            "reasoning_effort": "medium",
            "name": ANY,
        }
    ]
    assert parent_display.child_display.finished_statuses == ["completed"]
    assert parent_session_usage.total_tokens == 6
    assert parent_session_usage.sub_agent_total_tokens == 6
    assert parent_turn_usage.total_tokens == 6
    assert parent_turn_usage.sub_agent_total_tokens == 6
    assert captured["excluded_tools"] == {"sub_agent"}
    assert "delegated sub-agent" in str(captured["system_prompt"])
    assert isinstance(captured["settings"], Settings)
    assert captured["settings"].model == expected_model
    assert captured["settings"].reasoning_effort == "medium"



"""Textual sub-agent display — renders child-agent output inside a collapsible block."""

from __future__ import annotations

from typing import Any

from textual.widgets import Static

from pbi_agent.models.messages import TokenUsage, WebSearchSource
from pbi_agent.session_store import MessageRecord
from pbi_agent.ui.display_protocol import DisplayProtocol, PendingToolGroup
from pbi_agent.ui.formatting import (
    REDACTED_THINKING_NOTICE,
    escape_markup_text,
    format_patch_tool_item,
    format_shell_tool_item,
    format_usage_summary,
    format_wait_seconds,
    resolve_reasoning_panel,
    route_function_result,
    shorten,
    status_markup,
    tool_group_class,
    tool_item_class,
)
from pbi_agent.ui.widgets import (
    AssistantMarkdown,
    ErrorMessage,
    NoticeMessage,
    ThinkingBlock,
    ThinkingContent,
    ToolGroup,
    ToolItem,
    UsageSummary,
    WaitingIndicator,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pbi_agent.ui.display import Display


class SubAgentDisplay(DisplayProtocol):
    """Display bridge for sub-agent output rendered inside a Textual container."""

    def __init__(
        self,
        *,
        parent: Display,
        task_instruction: str,
        reasoning_effort: str | None,
        name: str = "sub_agent",
    ) -> None:
        self.parent = parent
        self.verbose = parent.verbose
        self._task_instruction = task_instruction
        self._reasoning_effort = reasoning_effort
        self._name = name
        self._tool_group = PendingToolGroup()
        self._waiting_widget_id: str | None = None
        self._active_thinking_widget_id: str | None = None
        self._counter = 0
        self._block_id = parent._next_id("subagent")
        self._body_id = f"{self._block_id}-body"
        self.parent._safe_call(
            self.parent.app.mount_sub_agent_block,
            self._block_id,
            self._title("running"),
            body_id=self._body_id,
        )

    def _title(self, status: str) -> str:
        summary = shorten(self._task_instruction.strip() or "task", 72)
        title = f"{self._name} \u00b7 {summary} \u00b7 {status}"
        if self._reasoning_effort:
            title += f" \u00b7 {self._reasoning_effort}"
        return title

    def _next_id(self, prefix: str) -> str:
        self._counter += 1
        return f"{self._block_id}-{prefix}-{self._counter}"

    def _mount_widget(self, widget: Static) -> None:
        self.parent._safe_call(
            self.parent.app.mount_widget_in_container,
            self._body_id,
            widget,
        )

    def _query_optional(self, selector: str) -> Any | None:
        return self.parent._safe_call(self.parent.app._query_optional, selector)

    def _mount_static_message(
        self,
        prefix: str,
        widget_cls: type[Static],
        text: str,
        *,
        classes: str = "",
    ) -> str:
        widget_id = self._next_id(prefix)
        kwargs: dict[str, Any] = {"id": widget_id}
        if classes:
            kwargs["classes"] = classes
        self._mount_widget(widget_cls(text, **kwargs))
        return widget_id

    def _start_tool_group(
        self,
        label: str,
        *,
        classes: str = "",
        function_count: int = 0,
    ) -> None:
        self._tool_group.start(label, classes=classes, function_count=function_count)

    def _append_tool_line(self, tool_name: str, text: str) -> None:
        self._tool_group.add_item(text, classes=tool_item_class(tool_name))

    # -- protocol stubs (no-ops for sub-agent context) -----------------------

    def request_shutdown(self) -> None:
        return None

    def submit_input(self, value: str) -> None:
        del value

    def request_new_chat(self) -> None:
        raise RuntimeError("Sub-agent display does not support interactive chat.")

    def reset_chat(self) -> None:
        return None

    def begin_sub_agent(
        self,
        *,
        task_instruction: str,
        reasoning_effort: str | None = None,
        name: str = "sub_agent",
    ) -> DisplayProtocol:
        del task_instruction, reasoning_effort, name
        return self

    def finish_sub_agent(self, *, status: str) -> None:
        self.wait_stop()
        self.parent._safe_call(
            self.parent.app.update_sub_agent_title,
            self._block_id,
            self._title(status),
        )

    def welcome(
        self,
        *,
        interactive: bool = True,
        model: str | None = None,
        reasoning_effort: str | None = None,
        single_turn_hint: str | None = None,
    ) -> None:
        del interactive, model, reasoning_effort, single_turn_hint

    def user_prompt(self) -> str:
        raise RuntimeError("Sub-agent display does not support user input.")

    def assistant_start(self) -> None:
        return None

    # -- rendering -----------------------------------------------------------

    def wait_start(self, message: str = "model is processing your request...") -> None:
        if self._waiting_widget_id is not None:
            return
        self._active_thinking_widget_id = None
        widget_id = self._next_id("wait")
        self._waiting_widget_id = widget_id
        self._mount_widget(WaitingIndicator(message=message, id=widget_id))

    def wait_stop(self) -> None:
        if self._waiting_widget_id is not None:
            self.parent._safe_call(
                self.parent.app.remove_widget, self._waiting_widget_id
            )
            self._waiting_widget_id = None

    def render_markdown(self, text: str) -> None:
        widget_id = self._next_id("md")
        self._mount_widget(AssistantMarkdown(text, id=widget_id))

    def render_thinking(
        self,
        text: str | None = None,
        *,
        title: str | None = None,
        replace_existing: bool = False,
        widget_id: str | None = None,
    ) -> str | None:
        summary = title or ""
        body, widget_title = resolve_reasoning_panel(text, summary)
        if body is None and not summary.strip():
            return None

        resolved_widget_id = widget_id
        if resolved_widget_id is None and replace_existing:
            resolved_widget_id = self._active_thinking_widget_id
        if resolved_widget_id is None:
            resolved_widget_id = self._next_id("thinking")
        if replace_existing:
            self._active_thinking_widget_id = resolved_widget_id

        existing = self._query_optional(f"#{resolved_widget_id}")
        if existing is None:
            self._mount_widget(
                ThinkingBlock(
                    ThinkingContent(body or "", id=f"{resolved_widget_id}-content"),
                    title=widget_title,
                    collapsed=True,
                    id=resolved_widget_id,
                )
            )
            return resolved_widget_id

        self.parent._safe_call(
            self.parent.app.update_thinking_block,
            resolved_widget_id,
            widget_title,
            body,
        )
        return resolved_widget_id

    def render_redacted_thinking(self) -> None:
        self._mount_static_message("redact", NoticeMessage, REDACTED_THINKING_NOTICE)

    def session_usage(self, usage: TokenUsage) -> None:
        del usage

    def turn_usage(self, usage: TokenUsage, elapsed_seconds: float) -> None:
        usage_text = format_usage_summary(
            usage.snapshot(),
            elapsed_seconds=elapsed_seconds,
            label=self._name,
        )
        self._mount_static_message("usage", UsageSummary, usage_text)

    # -- tool display --------------------------------------------------------

    def shell_start(self, commands: list[str]) -> None:
        self._start_tool_group(
            f"Running {len(commands)} shell command{'s' if len(commands) != 1 else ''}",
            classes=tool_group_class("shell"),
        )

    def shell_command(
        self,
        command: str,
        exit_code: int | None,
        timed_out: bool,
        *,
        call_id: str = "",
        working_directory: str = ".",
        timeout_ms: int | str = "default",
    ) -> None:
        self._append_tool_line(
            "shell",
            format_shell_tool_item(
                command,
                verbose=self.verbose,
                status=status_markup(timed_out=timed_out, exit_code=exit_code),
                call_id=call_id,
                working_directory=working_directory,
                timeout_ms=timeout_ms,
            ),
        )

    def patch_start(self, count: int) -> None:
        self._start_tool_group(
            f"Editing {count} file{'s' if count != 1 else ''}",
            classes=tool_group_class("apply_patch"),
        )

    def patch_result(
        self,
        path: str,
        operation: str,
        success: bool,
        *,
        call_id: str = "",
        detail: str = "",
    ) -> None:
        self._append_tool_line(
            "apply_patch",
            format_patch_tool_item(
                path,
                operation,
                verbose=self.verbose,
                status=status_markup(success=success),
                call_id=call_id,
                detail=detail,
            ),
        )

    def function_start(self, count: int) -> None:
        self._start_tool_group(
            f"Tool call{'s' if count != 1 else ''}",
            classes="tool-group-generic",
            function_count=count,
        )

    def function_result(
        self,
        name: str,
        success: bool,
        *,
        call_id: str = "",
        arguments: Any = None,
    ) -> None:
        self._tool_group.update_for_function(name)
        tool_name, text = route_function_result(
            name,
            verbose=self.verbose,
            status=status_markup(success=success),
            call_id=call_id,
            arguments=arguments,
        )
        self._append_tool_line(tool_name, text)

    def tool_group_end(self) -> None:
        if not self._tool_group.items:
            self._tool_group.reset()
            return
        group_id = self._next_id("tg")
        group = ToolGroup(
            *[
                ToolItem(item.text, classes=item.classes)
                for item in self._tool_group.items
            ],
            title=self._tool_group.label,
            collapsed=True,
            id=group_id,
            classes=self._tool_group.classes,
        )
        self._mount_widget(group)
        self._tool_group.reset()

    # -- notices -------------------------------------------------------------

    def retry_notice(self, attempt: int, max_retries: int) -> None:
        self.wait_stop()
        self._mount_static_message(
            "notice",
            NoticeMessage,
            f"Retrying\u2026 ({attempt}/{max_retries})",
        )

    def rate_limit_notice(
        self,
        *,
        wait_seconds: float,
        attempt: int,
        max_retries: int,
    ) -> None:
        self.wait_stop()
        self._mount_static_message(
            "notice",
            NoticeMessage,
            f"Rate limit reached. Retrying in {format_wait_seconds(wait_seconds)}s ({attempt}/{max_retries})",
        )

    def overload_notice(
        self,
        *,
        wait_seconds: float,
        attempt: int,
        max_retries: int,
    ) -> None:
        self.wait_stop()
        self._mount_static_message(
            "notice",
            NoticeMessage,
            f"Provider overloaded. Retrying in {format_wait_seconds(wait_seconds)}s ({attempt}/{max_retries})",
        )

    def error(self, message: str) -> None:
        self.wait_stop()
        self._tool_group.reset()
        self._mount_static_message(
            "err",
            ErrorMessage,
            f"Error: {escape_markup_text(message)}",
        )

    def debug(self, message: str) -> None:
        if self.verbose:
            self._mount_static_message(
                "dbg",
                Static,
                f"[dim]{message}[/dim]",
                classes="debug-msg",
            )

    def web_search_sources(self, sources: list[WebSearchSource]) -> None:
        del sources

    def replay_history(self, messages: list[MessageRecord]) -> None:
        del messages



from __future__ import annotations

import sys
from pathlib import Path

SYSTEM_PROMPT = """
You are pbi-agent, a local CLI coding agent for creating, auditing, and editing Power BI PBIP projects.

<environment>
- You run locally with workspace read/write access through function tools.
</environment>

<power_bi_rules>
- Always use explicit Power BI measures for displayed values in visuals; do not rely on implicit aggregations.
- Never create or rename a table to `Measures`; use `_Measures` for a dedicated measures table.
- Never modify auto-generated date tables whose names start with `DateTableTemplate_` or `LocalDateTable_`.
- When adding descriptions, skip those auto-generated date tables because their restricted TMDL schema does not support normal metadata edits.
- When laying out visuals, distribute them intentionally across the canvas and avoid clustering them into one area unless the user asked for that layout.
- If the user does not specify styling, apply the default preset for the visual type from the skill knowledge base.
- Style priority is: explicit user instruction > existing project or brand conventions > skill default preset.
</power_bi_rules>
""".strip()

_SUB_AGENT_PROMPT = """
<persona>
- You are a delegated sub-agent operating on behalf of the main agent.
- You are in background mode and will not interact with the user directly. Do not ask the user questions.
</persona>
""".strip()

SUB_AGENT_SYSTEM_PROMPT = f"{SYSTEM_PROMPT}\n\n{_SUB_AGENT_PROMPT}"

_MAX_PROJECT_RULES_BYTES = 1_000_000  # 1 MB


def _warn_project_rules(message: str) -> None:
    print(message, file=sys.stderr)


def load_project_rules(cwd: Path | None = None) -> str | None:
    """Read an optional ``AGENTS.md`` file from *cwd* (default: CWD).

    Returns the file content, or ``None`` when the file is absent, empty,
    or unreadable.
    """
    target = (cwd or Path.cwd()) / "AGENTS.md"

    try:
        size = target.stat().st_size
    except FileNotFoundError:
        return None
    except OSError:
        _warn_project_rules("AGENTS.md found but unreadable due to permissions.")
        return None

    if size > _MAX_PROJECT_RULES_BYTES:
        _warn_project_rules("AGENTS.md exceeds 1 MB; content will be truncated.")

    try:
        with target.open("rb") as fh:
            raw_content = fh.read(_MAX_PROJECT_RULES_BYTES)
    except OSError:
        _warn_project_rules("AGENTS.md found but unreadable due to permissions.")
        return None

    content = raw_content.decode("utf-8", errors="replace").strip()
    if not content:
        return None

    return content


def _append_project_rules(base_prompt: str) -> str:
    """Append ``<project_rules>`` section if ``AGENTS.md`` is present."""
    rules = load_project_rules()
    if rules is None:
        return base_prompt
    return f"{base_prompt}\n\n<project_rules>\n{rules}\n</project_rules>"


def get_system_prompt() -> str:
    return _append_project_rules(SYSTEM_PROMPT)


def get_sub_agent_system_prompt() -> str:
    return _append_project_rules(SUB_AGENT_SYSTEM_PROMPT)



"""Custom ``apply_patch`` tool – applies V4A diffs to workspace files.

This replaces the provider-specific native patch/editor tools (OpenAI
``apply_patch``, Anthropic ``str_replace_based_edit_tool``) with a single,
provider-agnostic function tool that goes through the normal tool registry
and execution pipeline.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

from pbi_agent.tools.apply_diff import apply_diff
from pbi_agent.tools.output import bound_output
from pbi_agent.tools.types import ToolContext, ToolSpec

SPEC = ToolSpec(
    name="apply_patch",
    description=(
        "Apply a file operation (create, update, or delete) using V4A diff format. "
        "Use this tool to create new files, edit existing files via diffs, "
        "or delete files in the workspace."
    ),
    parameters_schema={
        "type": "object",
        "properties": {
            "operation_type": {
                "type": "string",
                "enum": ["create_file", "update_file", "delete_file"],
                "description": "The type of file operation to perform.",
            },
            "path": {
                "type": "string",
                "description": (
                    "File path relative to the workspace root "
                    "(or absolute within workspace)."
                ),
            },
            "diff": {
                "type": "string",
                "description": (
                    "The V4A diff content. Required for create_file and "
                    "update_file operations. For create_file, lines must be "
                    "prefixed with '+'. For update_file, use context lines "
                    "(prefixed ' '), deletion lines ('-'), and insertion "
                    "lines ('+')."
                ),
            },
        },
        "required": ["operation_type", "path"],
        "additionalProperties": False,
    },
    is_destructive=True,
)


def handle(arguments: dict[str, Any], context: ToolContext) -> dict[str, Any]:
    """Execute a single file operation (create, update, or delete)."""
    operation_type = arguments.get("operation_type", "")
    path_value = arguments.get("path", "")
    diff = arguments.get("diff")

    if not isinstance(operation_type, str) or not operation_type:
        return {"error": "'operation_type' must be a non-empty string."}
    if not isinstance(path_value, str) or not path_value.strip():
        return {"error": "'path' must be a non-empty string."}

    root = Path.cwd().resolve()

    try:
        target_path = _resolve_safe_path(root, path_value)

        if operation_type == "create_file":
            _create_file(target_path, diff)
        elif operation_type == "update_file":
            _update_file(target_path, diff)
        elif operation_type == "delete_file":
            _delete_file(target_path)
        else:
            return {"error": f"Unsupported operation_type '{operation_type}'."}

        return {
            "status": "completed",
            "message": f"{operation_type} succeeded for '{path_value}'",
        }
    except Exception as exc:
        return {
            "status": "failed",
            "error": bound_output(str(exc))[0],
        }


# ---------------------------------------------------------------------------
# File operations
# ---------------------------------------------------------------------------


def _create_file(path: Path, diff: str | None) -> None:
    if path.exists():
        raise FileExistsError(f"file already exists: {path}")
    if not isinstance(diff, str) or not diff:
        raise ValueError("'diff' is required for create_file and must be non-empty.")
    content = apply_diff("", diff, mode="create")
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")


def _update_file(path: Path, diff: str | None) -> None:
    if not path.exists():
        raise FileNotFoundError(f"file not found: {path}")
    if not isinstance(diff, str) or not diff:
        raise ValueError("'diff' is required for update_file and must be non-empty.")
    current = path.read_text(encoding="utf-8")
    updated = apply_diff(current, diff, mode="default")
    path.write_text(updated, encoding="utf-8")


def _delete_file(path: Path) -> None:
    if not path.exists():
        raise FileNotFoundError(f"file not found: {path}")
    if path.is_dir():
        raise IsADirectoryError(f"path is a directory: {path}")
    path.unlink()


# ---------------------------------------------------------------------------
# Path safety
# ---------------------------------------------------------------------------


def _resolve_safe_path(root: Path, raw_path: str) -> Path:
    candidate = Path(raw_path)
    if candidate.is_absolute():
        resolved = candidate.resolve()
    else:
        resolved = (root / candidate).resolve()

    try:
        resolved.relative_to(root)
    except ValueError as exc:
        raise ValueError(f"path outside workspace is not allowed: {raw_path}") from exc

    return resolved



from __future__ import annotations

import json
import sys
from pathlib import Path

from pbi_agent.tools import python_exec as python_exec_tool
from pbi_agent.tools.types import ToolContext


def test_python_exec_runs_code_in_same_environment_and_workspace(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    monkeypatch.setenv("PBI_AGENT_TEST_FLAG", "present")
    (tmp_path / "helper_module.py").write_text(
        "VALUE = 'workspace import works'\n",
        encoding="utf-8",
    )

    result = python_exec_tool.handle(
        {
            "code": (
                "import json\n"
                "import os\n"
                "import sys\n"
                "from pathlib import Path\n"
                "from helper_module import VALUE\n"
                "payload = {\n"
                "    'cwd': str(Path.cwd()),\n"
                "    'env': os.environ['PBI_AGENT_TEST_FLAG'],\n"
                "    'executable': sys.executable,\n"
                "    'value': VALUE,\n"
                "}\n"
                "print(json.dumps(payload, sort_keys=True))\n"
            )
        },
        ToolContext(),
    )

    assert result["ok"] is True
    assert result["stderr"] == ""
    assert result["error_type"] is None
    assert result["timed_out"] is False
    payload = json.loads(result["stdout"])
    assert payload == {
        "cwd": str(tmp_path.resolve()),
        "env": "present",
        "executable": sys.executable,
        "value": "workspace import works",
    }


def test_python_exec_captures_structured_result(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)

    result = python_exec_tool.handle(
        {
            "code": "result = {'status': 'ok', 'value': 42}",
            "capture_result": True,
        },
        ToolContext(),
    )

    assert result == {
        "ok": True,
        "stdout": "",
        "stderr": "",
        "result": {"status": "ok", "value": 42},
        "error_type": None,
        "error_message": None,
        "timed_out": False,
        "execution_time_ms": result["execution_time_ms"],
        "stdout_truncated": False,
        "stderr_truncated": False,
        "result_truncated": False,
    }
    assert result["execution_time_ms"] >= 0


def test_python_exec_reports_python_exceptions(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)

    result = python_exec_tool.handle(
        {"code": "raise ValueError('bad input')"},
        ToolContext(),
    )

    assert result["ok"] is False
    assert result["error_type"] == "ValueError"
    assert result["error_message"] == "bad input"
    assert "Traceback" in result["stderr"]
    assert "ValueError: bad input" in result["stderr"]
    assert result["timed_out"] is False


def test_python_exec_enforces_timeout(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)

    result = python_exec_tool.handle(
        {
            "code": "import time\ntime.sleep(2)",
            "timeout_seconds": 1,
        },
        ToolContext(),
    )

    assert result["ok"] is False
    assert result["error_type"] == "TimeoutError"
    assert result["error_message"] == "Execution exceeded timeout of 1 second"
    assert result["timed_out"] is True
    assert result["result"] is None


def test_python_exec_bounds_large_stdout(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)

    result = python_exec_tool.handle(
        {"code": "print('A' * 50000)"},
        ToolContext(),
    )

    assert result["ok"] is True
    assert result["stdout_truncated"] is True
    assert len(result["stdout"]) <= python_exec_tool.MAX_STDOUT_CHARS
    assert "chars omitted" in result["stdout"]


def test_python_exec_reports_non_serializable_result(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.chdir(tmp_path)

    result = python_exec_tool.handle(
        {
            "code": "result = {1, 2, 3}",
            "capture_result": True,
        },
        ToolContext(),
    )

    assert result["ok"] is True
    assert result["result"] is None
    assert "Failed to serialize `result` as JSON" in result["stderr"]
    assert result["error_type"] is None


def test_python_exec_treats_zero_system_exit_as_success(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.chdir(tmp_path)

    result = python_exec_tool.handle(
        {
            "code": "import sys\nsys.exit(0)",
            "capture_result": True,
        },
        ToolContext(),
    )

    assert result["ok"] is True
    assert result["stderr"] == ""
    assert result["error_type"] is None
    assert result["error_message"] is None
    assert result["result"] is None


def test_python_exec_truncates_large_structured_result_in_place(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.chdir(tmp_path)

    result = python_exec_tool.handle(
        {
            "code": (
                "result = {\n"
                "    'status': 'ok',\n"
                f"    'payload': 'x' * {python_exec_tool.MAX_RESULT_CHARS * 2},\n"
                "}\n"
            ),
            "capture_result": True,
        },
        ToolContext(),
    )

    assert result["ok"] is True
    assert result["result_truncated"] is True
    assert isinstance(result["result"], dict)
    assert result["result"]["status"] == "ok"
    assert isinstance(result["result"]["payload"], str)
    assert len(json.dumps(result["result"])) <= python_exec_tool.MAX_RESULT_CHARS


def test_python_exec_bounds_response_file_before_parent_reads_it(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.chdir(tmp_path)
    captured_temp_path: Path | None = None

    class PersistentTemporaryDirectory:
        def __enter__(self) -> str:
            nonlocal captured_temp_path
            captured_temp_path = tmp_path / "python-exec-temp"
            captured_temp_path.mkdir()
            return str(captured_temp_path)

        def __exit__(self, exc_type, exc, tb) -> bool:
            return False

    monkeypatch.setattr(
        python_exec_tool.tempfile,
        "TemporaryDirectory",
        lambda prefix: PersistentTemporaryDirectory(),
    )

    result = python_exec_tool.handle(
        {
            "code": f"result = 'x' * {python_exec_tool.MAX_RESULT_CHARS * 4}",
            "capture_result": True,
        },
        ToolContext(),
    )

    assert result["ok"] is True
    assert result["result_truncated"] is True
    assert captured_temp_path is not None

    response_path = captured_temp_path / "response.json"
    response_payload = json.loads(response_path.read_text(encoding="utf-8"))
    assert response_payload["result_truncated"] is True
    assert response_path.stat().st_size <= python_exec_tool.MAX_RESULT_CHARS + 1024


def test_python_exec_uses_requested_working_directory(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.chdir(tmp_path)
    nested = tmp_path / "nested"
    nested.mkdir()
    (nested / "marker.txt").write_text("hello", encoding="utf-8")

    result = python_exec_tool.handle(
        {
            "code": (
                "from pathlib import Path\n"
                "result = {\n"
                "    'cwd': str(Path.cwd()),\n"
                "    'marker': Path('marker.txt').read_text(encoding='utf-8'),\n"
                "}\n"
            ),
            "working_directory": "nested",
            "capture_result": True,
        },
        ToolContext(),
    )

    assert result["ok"] is True
    assert result["result"] == {
        "cwd": str(nested.resolve()),
        "marker": "hello",
    }


def test_python_exec_rejects_invalid_working_directory(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    outside = tmp_path.parent

    result = python_exec_tool.handle(
        {"code": "print('hi')", "working_directory": str(outside)},
        ToolContext(),
    )

    assert result["ok"] is False
    assert result["error_type"] == "ValueError"
    assert "outside workspace" in result["error_message"]



from __future__ import annotations

import urllib.error
import urllib.request

from pbi_agent.tools import read_web_url as read_web_url_tool
from pbi_agent.tools.types import ToolContext


class _FakeResponse:
    def __init__(self, body: str, *, headers: dict[str, str] | None = None) -> None:
        self._body = body.encode("utf-8")
        self.headers = headers or {}

    def read(self) -> bytes:
        return self._body

    def __enter__(self) -> "_FakeResponse":
        return self

    def __exit__(self, *_: object) -> None:
        return None


def test_read_web_url_returns_markdown(monkeypatch) -> None:
    seen: dict[str, object] = {}

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ) -> _FakeResponse:
        seen["url"] = request.full_url
        seen["method"] = request.get_method()
        seen["accept"] = request.get_header("Accept")
        seen["content_type"] = request.headers.get("Content-type")
        seen["body"] = request.data.decode("utf-8")
        seen["timeout"] = timeout
        return _FakeResponse("# Example\n\nBody", headers={"x-markdown-tokens": "725"})

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    result = read_web_url_tool.handle(
        {"url": "https://example.com/docs"},
        ToolContext(),
    )

    assert result == {
        "url": "https://example.com/docs",
        "markdown": "# Example\n\nBody",
    }
    assert seen == {
        "url": read_web_url_tool.MARKDOWN_NEW_URL,
        "method": "POST",
        "accept": "text/markdown",
        "content_type": "application/json",
        "body": '{"url": "https://example.com/docs"}',
        "timeout": read_web_url_tool._REQUEST_TIMEOUT_SECS,
    }


def test_read_web_url_bounds_large_markdown(monkeypatch) -> None:
    long_markdown = (
        f"# Start\n\nprefix-{'x' * (read_web_url_tool.MAX_MARKDOWN_CHARS + 200)}-suffix"
    )

    monkeypatch.setattr(
        urllib.request,
        "urlopen",
        lambda request, timeout: _FakeResponse(long_markdown),
    )

    result = read_web_url_tool.handle({"url": "https://example.com"}, ToolContext())

    assert result["url"] == "https://example.com"
    assert result["markdown_truncated"] is True
    assert len(result["markdown"]) <= read_web_url_tool.MAX_MARKDOWN_CHARS
    assert result["markdown"].startswith("# Start\n\nprefix-")
    assert result["markdown"].endswith("-suffix")
    assert "chars omitted" in result["markdown"]


def test_read_web_url_rejects_invalid_urls_before_network(monkeypatch) -> None:
    called = False

    def fake_urlopen(request: urllib.request.Request, timeout: float) -> _FakeResponse:
        del request, timeout
        nonlocal called
        called = True
        return _FakeResponse("unused")

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    result = read_web_url_tool.handle({"url": "ftp://example.com"}, ToolContext())

    assert result == {"error": "'url' must be an absolute http or https URL."}
    assert called is False


def test_read_web_url_reports_http_errors(monkeypatch) -> None:
    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ) -> _FakeResponse:
        del request, timeout
        raise urllib.error.HTTPError(
            read_web_url_tool.MARKDOWN_NEW_URL,
            429,
            "Too Many Requests",
            hdrs=None,
            fp=None,
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    result = read_web_url_tool.handle({"url": "https://example.com"}, ToolContext())

    assert result == {"error": "HTTP 429 Too Many Requests"}


def test_read_web_url_reports_network_failures(monkeypatch) -> None:
    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ) -> _FakeResponse:
        del request, timeout
        raise urllib.error.URLError("temporary failure")

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    result = read_web_url_tool.handle({"url": "https://example.com"}, ToolContext())

    assert result == {"error": "failed to fetch URL: temporary failure"}



from __future__ import annotations

import pytest

from unittest.mock import patch

from pbi_agent.web import chat_entry


def test_run_starts_watchdog_and_launches_console() -> None:
    with (
        patch("pbi_agent.web.chat_entry._start_parent_watchdog") as mock_watchdog,
        patch("pbi_agent.web.chat_entry.main", return_value=7) as mock_main,
    ):
        rc = chat_entry.run(["--parent-pid", "4321", "--verbose"])

    assert rc == 7
    mock_watchdog.assert_called_once_with(4321)
    mock_main.assert_called_once_with(["--verbose", "console"])


def test_watch_parent_process_uses_pid_existence_check_on_posix() -> None:
    with (
        patch.object(chat_entry.os, "name", "posix"),
        patch(
            "pbi_agent.web.chat_entry.os.getppid",
            side_effect=AssertionError("getppid should not be used on posix"),
        ),
        patch(
            "pbi_agent.web.chat_entry.os.kill",
            side_effect=ProcessLookupError,
        ) as mock_kill,
        patch(
            "pbi_agent.web.chat_entry.os._exit",
            side_effect=SystemExit(0),
        ) as mock_exit,
    ):
        with pytest.raises(SystemExit):
            chat_entry._watch_parent_process(4321)

    mock_kill.assert_called_once_with(4321, 0)
    mock_exit.assert_called_once_with(0)



from __future__ import annotations

from pbi_agent.tools.output import MAX_OUTPUT_CHARS, bound_output


def test_bound_output_preserves_start_and_end_when_truncated() -> None:
    original = f"start-{'x' * (MAX_OUTPUT_CHARS + 200)}-end"

    bounded, truncated = bound_output(original)

    assert truncated is True
    assert len(bounded) <= MAX_OUTPUT_CHARS
    assert bounded.startswith("start-")
    assert bounded.endswith("-end")
    assert "chars omitted" in bounded


def test_bound_output_falls_back_to_simple_ellipsis_for_small_limits() -> None:
    bounded, truncated = bound_output("abcdef", limit=4)

    assert truncated is True
    assert bounded == "abc…"



from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Callable, Iterator

from pbi_agent.tools.output import bound_output
from pbi_agent.tools.types import ToolContext, ToolSpec
from pbi_agent.tools.workspace_filters import build_glob_matcher
from pbi_agent.tools.workspace_filters import should_skip_directory_name
from pbi_agent.tools.workspace_access import DEFAULT_MAX_ENTRIES
from pbi_agent.tools.workspace_access import normalize_positive_int
from pbi_agent.tools.workspace_access import relative_workspace_path
from pbi_agent.tools.workspace_access import resolve_safe_path

SPEC = ToolSpec(
    name="list_files",
    description=(
        "List files and directories within the workspace, with optional glob and "
        "type filtering for targeted filename or path lookups. Recursive listings "
        "skip common generated and dependency directories."
    ),
    parameters_schema={
        "type": "object",
        "properties": {
            "path": {
                "type": "string",
                "description": (
                    "Directory or file path relative to the workspace root "
                    "(or absolute within workspace). Defaults to '.'."
                ),
            },
            "recursive": {
                "type": "boolean",
                "description": "Whether to traverse subdirectories recursively.",
                "default": True,
            },
            "glob": {
                "type": "string",
                "description": (
                    "Optional glob filter. Match against the entry name unless "
                    "the pattern includes a path separator."
                ),
            },
            "entry_type": {
                "type": "string",
                "description": (
                    "Limit results to files, directories, or both. Defaults to 'all'."
                ),
                "enum": ["all", "file", "directory"],
                "default": "all",
            },
            "max_entries": {
                "type": "integer",
                "description": "Maximum number of entries to return. Defaults to 200.",
            },
        },
        "additionalProperties": False,
    },
)


def handle(arguments: dict[str, Any], context: ToolContext) -> dict[str, Any]:
    del context
    root = Path.cwd().resolve()

    try:
        target_path = resolve_safe_path(root, arguments.get("path"))
        recursive = arguments.get("recursive", True)
        glob_pattern = _normalize_glob_pattern(arguments.get("glob"))
        entry_type = _normalize_entry_type(arguments.get("entry_type"))
        if entry_type is None:
            return {"error": "'entry_type' must be one of: all, file, directory."}
        matcher = build_glob_matcher(glob_pattern)
        max_entries = normalize_positive_int(
            arguments.get("max_entries"),
            default=DEFAULT_MAX_ENTRIES,
        )

        if not target_path.exists():
            return {"error": f"path not found: {target_path}"}

        if target_path.is_file():
            relative_path = relative_workspace_path(root, target_path)
            entry = _build_entry(relative_path, "file")
            entries: list[dict[str, Any]] = []
            if _matches_filters(
                relative_path, target_path.name, "file", matcher, entry_type
            ):
                entry, path_truncated = _bound_entry_path(entry)
                if path_truncated:
                    entry["path_truncated"] = True
                entries.append(entry)
            return {
                "entries": entries,
                "returned_entries": len(entries),
                "total_entries": len(entries),
                "has_more": False,
            }

        if not target_path.is_dir():
            return {"error": f"path is not a regular file or directory: {target_path}"}

        matching_entries: list[dict[str, Any]] = []
        entries_truncated = False
        for relative_path, name, candidate_type in _iter_entries(
            root,
            target_path,
            recursive=bool(recursive),
        ):
            if not _matches_filters(
                relative_path,
                name,
                candidate_type,
                matcher,
                entry_type,
            ):
                continue
            if len(matching_entries) >= max_entries:
                entries_truncated = True
                break
            entry = _build_entry(relative_path, candidate_type)
            bounded_entry, path_truncated = _bound_entry_path(entry)
            if path_truncated:
                bounded_entry["path_truncated"] = True
            matching_entries.append(bounded_entry)

        result: dict[str, Any] = {
            "entries": matching_entries,
            "returned_entries": len(matching_entries),
            "has_more": entries_truncated,
        }
        if entries_truncated:
            result["entries_truncated"] = True
        else:
            result["total_entries"] = len(matching_entries)
        return result
    except Exception as exc:
        return {"error": bound_output(str(exc))[0]}


def _build_entry(relative_path: str, entry_type: str) -> dict[str, Any]:
    return {
        "path": relative_path,
        "type": entry_type,
    }


def _bound_entry_path(entry: dict[str, Any]) -> tuple[dict[str, Any], bool]:
    bounded_path, truncated = bound_output(str(entry["path"]))
    return {
        **entry,
        "path": bounded_path,
    }, truncated


def _normalize_glob_pattern(raw_value: Any) -> str | None:
    if not isinstance(raw_value, str):
        return None
    normalized = raw_value.strip()
    return normalized or None


def _normalize_entry_type(raw_value: Any) -> str | None:
    if raw_value is None:
        return "all"
    if not isinstance(raw_value, str):
        return None
    normalized = raw_value.strip().lower()
    if not normalized:
        return "all"
    if normalized in {"all", "file", "directory"}:
        return normalized
    return None


def _matches_filters(
    relative_path: str,
    name: str,
    candidate_type: str,
    matcher: Callable[[str, str], bool],
    entry_type: str,
) -> bool:
    if entry_type == "file" and candidate_type != "file":
        return False
    if entry_type == "directory" and candidate_type != "directory":
        return False
    return matcher(relative_path, name)


def _iter_entries(
    root: Path,
    target_path: Path,
    *,
    recursive: bool,
) -> Iterator[tuple[str, str, str]]:
    if not recursive:
        with os.scandir(target_path) as scan:
            entries = sorted(scan, key=lambda entry: entry.name.casefold())
            for entry in entries:
                candidate_type = _entry_type_from_dir_entry(entry)
                if candidate_type is None:
                    continue
                relative_path = (target_path / entry.name).relative_to(root).as_posix()
                yield relative_path, entry.name, candidate_type
        return

    for current_root, dirnames, filenames in os.walk(
        target_path,
        topdown=True,
        followlinks=False,
    ):
        dirnames[:] = [
            dirname
            for dirname in sorted(dirnames, key=str.casefold)
            if not should_skip_directory_name(dirname)
        ]
        filenames.sort(key=str.casefold)
        current = Path(current_root)

        for dirname in dirnames:
            relative_path = (current / dirname).relative_to(root).as_posix()
            yield relative_path, dirname, "directory"

        for filename in filenames:
            relative_path = (current / filename).relative_to(root).as_posix()
            yield relative_path, filename, "file"


def _entry_type_from_dir_entry(entry: os.DirEntry[str]) -> str | None:
    if entry.is_dir(follow_symlinks=False):
        return "directory"
    if entry.is_file(follow_symlinks=False):
        return "file"
    return None



"""Custom ``shell`` tool – executes shell commands in a subprocess.

This replaces the provider-specific native shell tools (OpenAI ``shell``,
Anthropic ``bash``) with a single, provider-agnostic function tool that goes
through the normal tool registry and execution pipeline.
"""

from __future__ import annotations

import subprocess
from pathlib import Path
from typing import Any

from pbi_agent.tools.output import MAX_OUTPUT_CHARS as DEFAULT_MAX_OUTPUT_CHARS
from pbi_agent.tools.output import bound_output, decode_output
from pbi_agent.tools.types import ToolContext, ToolSpec

MAX_TIMEOUT_MS = 120_000
MAX_OUTPUT_CHARS = DEFAULT_MAX_OUTPUT_CHARS

SPEC = ToolSpec(
    name="shell",
    description=(
        "Execute a shell command and return its stdout, stderr, and exit code. "
        "Commands run inside the workspace directory by default. "
        "Use this for running builds, tests, git, file listing, or any CLI tool."
    ),
    parameters_schema={
        "type": "object",
        "properties": {
            "command": {
                "type": "string",
                "description": "The shell command to execute.",
            },
            "working_directory": {
                "type": "string",
                "description": (
                    "Working directory for the command, relative to the "
                    "workspace root. Defaults to the workspace root."
                ),
            },
            "timeout_ms": {
                "type": "integer",
                "description": (
                    "Timeout in milliseconds (max 120 000). "
                    "Defaults to 120 000 (2 minutes)."
                ),
            },
        },
        "required": ["command"],
        "additionalProperties": False,
    },
    is_destructive=True,
)


def handle(arguments: dict[str, Any], context: ToolContext) -> dict[str, Any]:
    """Execute a single shell command and return structured output."""
    command = arguments.get("command", "")
    if not isinstance(command, str) or not command.strip():
        return {"error": "'command' must be a non-empty string."}

    root = Path.cwd().resolve()
    working_directory = _resolve_working_directory(
        root, arguments.get("working_directory")
    )
    timeout_ms = _normalize_timeout_ms(arguments.get("timeout_ms"))

    try:
        completed = subprocess.run(
            command,
            cwd=str(working_directory),
            capture_output=True,
            text=False,
            shell=True,
            timeout=(timeout_ms / 1000.0),
        )
        return {
            **_build_output_payload(
                stdout=decode_output(completed.stdout),
                stderr=decode_output(completed.stderr),
            ),
            "exit_code": completed.returncode,
        }
    except subprocess.TimeoutExpired as exc:
        return {
            **_build_output_payload(
                stdout=decode_output(exc.stdout),
                stderr=decode_output(exc.stderr),
            ),
            "exit_code": None,
            "timed_out": True,
            "error": f"Command timed out after {timeout_ms}ms.",
        }
    except Exception as exc:
        stderr, stderr_truncated = bound_output(str(exc))
        return {
            "stdout": "",
            "stderr": stderr,
            "exit_code": 1,
            "error": f"Shell execution failed: {exc}",
            **({"stderr_truncated": True} if stderr_truncated else {}),
        }


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _resolve_working_directory(root: Path, raw: Any) -> Path:
    if raw is None:
        return root
    if not isinstance(raw, str) or not raw.strip():
        return root

    candidate = Path(raw)
    if candidate.is_absolute():
        resolved = candidate.resolve()
    else:
        resolved = (root / candidate).resolve()

    try:
        resolved.relative_to(root)
    except ValueError as exc:
        raise ValueError(
            f"working_directory outside workspace is not allowed: {raw}"
        ) from exc

    if not resolved.exists() or not resolved.is_dir():
        raise FileNotFoundError(f"working_directory does not exist: {resolved}")
    return resolved


def _normalize_timeout_ms(raw_timeout: Any) -> int:
    if raw_timeout is None:
        return MAX_TIMEOUT_MS
    if not isinstance(raw_timeout, int):
        return MAX_TIMEOUT_MS
    if raw_timeout < 1:
        return MAX_TIMEOUT_MS
    return min(raw_timeout, MAX_TIMEOUT_MS)


def _build_output_payload(*, stdout: str, stderr: str) -> dict[str, Any]:
    bounded_stdout, stdout_truncated = bound_output(stdout)
    bounded_stderr, stderr_truncated = bound_output(stderr)
    payload: dict[str, Any] = {
        "stdout": bounded_stdout,
        "stderr": bounded_stderr,
    }
    if stdout_truncated:
        payload["stdout_truncated"] = True
    if stderr_truncated:
        payload["stderr_truncated"] = True
    return payload



from __future__ import annotations

import argparse
import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from dotenv import load_dotenv

DEFAULT_RESPONSES_URL = "https://api.openai.com/v1/responses"
DEFAULT_XAI_RESPONSES_URL = "https://api.x.ai/v1/responses"
DEFAULT_GOOGLE_INTERACTIONS_URL = (
    "https://generativelanguage.googleapis.com/v1beta/interactions"
)
DEFAULT_GENERIC_API_URL = "https://openrouter.ai/api/v1/chat/completions"
DEFAULT_MODEL = "gpt-5.4-2026-03-05"
DEFAULT_XAI_MODEL = "grok-4-1-fast-reasoning"
DEFAULT_GOOGLE_MODEL = "gemini-3.1-flash-lite-preview"
DEFAULT_ANTHROPIC_MODEL = "claude-opus-4-6"
DEFAULT_MAX_TOKENS = 16384
OPENAI_SERVICE_TIERS = ("auto", "default", "flex", "priority")
PROVIDER_API_KEY_ENVS = {
    "openai": "OPENAI_API_KEY",
    "xai": "XAI_API_KEY",
    "google": "GEMINI_API_KEY",
    "anthropic": "ANTHROPIC_API_KEY",
    "generic": "GENERIC_API_KEY",
}
INTERNAL_CONFIG_PATH_ENV = "PBI_AGENT_INTERNAL_CONFIG_PATH"
DEFAULT_INTERNAL_CONFIG_PATH = Path.home() / ".pbi-agent" / "config.json"


class ConfigError(ValueError):
    """Raised when required runtime configuration is invalid."""


def missing_api_key_message(provider: str) -> str:
    if provider == "google":
        return (
            "Missing API key for provider 'google'. Set GEMINI_API_KEY (or "
            "PBI_AGENT_API_KEY) in environment, or pass --google-api-key "
            "(or --api-key)."
        )
    return (
        f"Missing API key for provider '{provider}'. Set PBI_AGENT_API_KEY in "
        "environment or pass --api-key."
    )


@dataclass(slots=True)
class Settings:
    api_key: str
    responses_url: str = DEFAULT_RESPONSES_URL
    model: str = DEFAULT_MODEL
    sub_agent_model: str | None = None
    max_tokens: int = DEFAULT_MAX_TOKENS
    verbose: bool = False
    max_tool_workers: int = 4
    max_retries: int = 3
    reasoning_effort: str = "xhigh"
    compact_threshold: int = 200000
    # Provider selection
    provider: str = "openai"
    generic_api_url: str = DEFAULT_GENERIC_API_URL
    service_tier: str | None = None
    web_search: bool = True

    def validate(self) -> None:
        if self.provider not in {"openai", "xai", "google", "anthropic", "generic"}:
            raise ConfigError(
                "--provider must be one of: openai, xai, google, anthropic, generic."
            )
        if not self.api_key:
            raise ConfigError(missing_api_key_message(self.provider))
        if self.max_tool_workers < 1:
            raise ConfigError("--max-tool-workers must be >= 1.")
        if self.max_retries < 0:
            raise ConfigError("--max-retries must be >= 0.")
        if self.reasoning_effort not in {"low", "medium", "high", "xhigh"}:
            raise ConfigError(
                "--reasoning-effort must be one of: low, medium, high, xhigh."
            )
        if self.compact_threshold < 1:
            raise ConfigError("--compact-threshold must be >= 1.")
        if self.max_tokens < 1:
            raise ConfigError("--max-tokens must be >= 1.")
        if self.service_tier is not None and self.provider != "openai":
            raise ConfigError(
                "--service-tier is only supported with the OpenAI provider."
            )
        if (
            self.service_tier is not None
            and self.service_tier not in OPENAI_SERVICE_TIERS
        ):
            allowed = ", ".join(OPENAI_SERVICE_TIERS)
            raise ConfigError(f"--service-tier must be one of: {allowed}.")

    def redacted(self) -> dict[str, str | int | bool | None]:
        return {
            "provider": self.provider,
            "api_key": redact_secret(self.api_key),
            "responses_url": self.responses_url,
            "model": self.model,
            "sub_agent_model": self.sub_agent_model,
            "max_tokens": self.max_tokens,
            "verbose": self.verbose,
            "max_tool_workers": self.max_tool_workers,
            "max_retries": self.max_retries,
            "reasoning_effort": self.reasoning_effort,
            "compact_threshold": self.compact_threshold,
            "generic_api_url": self.generic_api_url,
            "service_tier": self.service_tier,
            "web_search": self.web_search,
        }


def redact_secret(value: str) -> str:
    if not value:
        return ""
    if len(value) <= 8:
        return "***"
    return f"{value[:4]}...{value[-4:]}"


def _default_responses_url(provider: str) -> str:
    if provider == "xai":
        return DEFAULT_XAI_RESPONSES_URL
    if provider == "google":
        return DEFAULT_GOOGLE_INTERACTIONS_URL
    return DEFAULT_RESPONSES_URL


def _default_model(provider: str) -> str:
    if provider == "generic":
        return ""
    if provider == "xai":
        return DEFAULT_XAI_MODEL
    if provider == "google":
        return DEFAULT_GOOGLE_MODEL
    if provider == "anthropic":
        return DEFAULT_ANTHROPIC_MODEL
    return DEFAULT_MODEL


def resolve_settings(args: argparse.Namespace) -> Settings:
    load_dotenv()

    internal_config = _read_internal_config()
    providers_config = _provider_configs(internal_config)

    provider_from_cli = getattr(args, "provider", None)
    provider_from_env = os.getenv("PBI_AGENT_PROVIDER")
    provider = (
        provider_from_cli
        or provider_from_env
        or _last_used_provider(internal_config)
        or "openai"
    )
    provider_config = providers_config.get(provider, {})

    api_key = (
        getattr(args, "api_key", None)
        or os.getenv("PBI_AGENT_API_KEY", "")
        or os.getenv(PROVIDER_API_KEY_ENVS.get(provider, ""), "")
        or str(provider_config.get("api_key", ""))
    )
    responses_url_override = getattr(args, "responses_url", None) or os.getenv(
        "PBI_AGENT_RESPONSES_URL"
    )
    generic_api_url = (
        getattr(args, "generic_api_url", None)
        or os.getenv("PBI_AGENT_GENERIC_API_URL")
        or _config_string(provider_config, "generic_api_url")
    )
    responses_url = (
        responses_url_override
        or _config_string(provider_config, "responses_url")
        or _default_responses_url(provider)
    )
    model_override = args.model or os.getenv("PBI_AGENT_MODEL")
    if not model_override:
        model_override = _config_string(provider_config, "model")
    model = model_override or _default_model(provider)
    sub_agent_model = (
        getattr(args, "sub_agent_model", None)
        or os.getenv("PBI_AGENT_SUB_AGENT_MODEL")
        or _config_string(provider_config, "sub_agent_model")
    )
    max_tool_workers = args.max_tool_workers
    if max_tool_workers is None:
        max_tool_workers = int(
            os.getenv(
                "PBI_AGENT_MAX_TOOL_WORKERS",
                str(_config_int(provider_config, "max_tool_workers", 4)),
            )
        )
    max_retries = args.max_retries
    if max_retries is None:
        max_retries = int(
            os.getenv(
                "PBI_AGENT_MAX_RETRIES",
                str(_config_int(provider_config, "max_retries", 3)),
            )
        )
    default_effort = "xhigh" if provider == "openai" else "high"
    reasoning_effort = (
        args.reasoning_effort
        or os.getenv("PBI_AGENT_REASONING_EFFORT")
        or _config_string(provider_config, "reasoning_effort")
        or default_effort
    )
    compact_threshold = args.compact_threshold
    if compact_threshold is None:
        compact_threshold = int(
            os.getenv(
                "PBI_AGENT_COMPACT_THRESHOLD",
                str(_config_int(provider_config, "compact_threshold", 150000)),
            )
        )

    max_tokens_raw = getattr(args, "max_tokens", None)

    if max_tokens_raw is None:
        max_tokens = int(
            os.getenv(
                "PBI_AGENT_MAX_TOKENS",
                str(_config_int(provider_config, "max_tokens", DEFAULT_MAX_TOKENS)),
            )
        )
    else:
        max_tokens = int(max_tokens_raw)

    service_tier = (
        getattr(args, "service_tier", None)
        or os.getenv("PBI_AGENT_SERVICE_TIER")
        or _config_string(provider_config, "service_tier")
    )

    no_web_search = getattr(args, "no_web_search", False)
    web_search_env = os.getenv("PBI_AGENT_WEB_SEARCH")
    if no_web_search:
        web_search = False
    elif web_search_env is not None:
        web_search = web_search_env.lower() not in ("0", "false", "no")
    else:
        web_search = _config_bool(provider_config, "web_search", True)

    return Settings(
        api_key=api_key,
        responses_url=responses_url,
        generic_api_url=generic_api_url or DEFAULT_GENERIC_API_URL,
        model=model,
        sub_agent_model=sub_agent_model,
        max_tokens=max_tokens,
        verbose=bool(args.verbose),
        max_tool_workers=max_tool_workers,
        max_retries=max_retries,
        reasoning_effort=reasoning_effort,
        compact_threshold=compact_threshold,
        provider=provider,
        service_tier=service_tier,
        web_search=web_search,
    )


def save_internal_config(settings: Settings) -> None:
    path = _internal_config_path()
    data = _read_internal_config()
    providers = _provider_configs(data)
    providers[settings.provider] = {
        "api_key": settings.api_key,
        "responses_url": settings.responses_url,
        "generic_api_url": settings.generic_api_url,
        "model": settings.model,
        "sub_agent_model": settings.sub_agent_model,
        "max_tokens": settings.max_tokens,
        "reasoning_effort": settings.reasoning_effort,
        "max_tool_workers": settings.max_tool_workers,
        "max_retries": settings.max_retries,
        "compact_threshold": settings.compact_threshold,
        "service_tier": settings.service_tier,
        "web_search": settings.web_search,
    }
    data["providers"] = providers
    data["last_used_provider"] = settings.provider
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(data, indent=2, sort_keys=True), encoding="utf-8")


def _internal_config_path() -> Path:
    configured_path = os.getenv(INTERNAL_CONFIG_PATH_ENV)
    if configured_path:
        return Path(configured_path).expanduser().resolve()
    return DEFAULT_INTERNAL_CONFIG_PATH


def _read_internal_config() -> dict[str, Any]:
    path = _internal_config_path()
    if not path.exists():
        return {}
    try:
        payload = json.loads(path.read_text(encoding="utf-8"))
    except (OSError, json.JSONDecodeError):
        return {}
    if not isinstance(payload, dict):
        return {}
    return payload


def _provider_configs(config: dict[str, Any]) -> dict[str, dict[str, Any]]:
    providers = config.get("providers")
    if not isinstance(providers, dict):
        return {}
    return {
        name: payload
        for name, payload in providers.items()
        if isinstance(name, str) and isinstance(payload, dict)
    }


def _last_used_provider(config: dict[str, Any]) -> str | None:
    last_used = config.get("last_used_provider")
    if isinstance(last_used, str):
        return last_used
    return None


def _config_string(provider_config: dict[str, Any], key: str) -> str | None:
    value = provider_config.get(key)
    if isinstance(value, str):
        return value
    return None


def _config_int(provider_config: dict[str, Any], key: str, fallback: int) -> int:
    value = provider_config.get(key)
    if isinstance(value, int):
        return value
    return fallback


def _config_bool(provider_config: dict[str, Any], key: str, fallback: bool) -> bool:
    value = provider_config.get(key)
    if isinstance(value, bool):
        return value
    return fallback



"""Prompt builder for ``pbi-agent audit`` mode."""

from __future__ import annotations

import shutil
from importlib.resources import as_file, files
from pathlib import Path

AUDIT_REPORT_FILENAME = "AUDIT-REPORT.md"
AUDIT_TODO_FILENAME = "AUDIT-TODO.md"


def copy_audit_todo(dest: Path) -> Path:
    """Copy the bundled ``AUDIT-TODO.md`` template into *dest*.

    If the file already exists it is left untouched so that a resumed audit
    can pick up where it left off.

    Returns
    -------
    Path
        The full path to the (existing or newly created) file.
    """
    target = dest / AUDIT_TODO_FILENAME
    if target.exists():
        return target
    source_traversable = files("pbi_agent.agent").joinpath(AUDIT_TODO_FILENAME)
    with as_file(source_traversable) as source_path:
        shutil.copy2(source_path, target)
    return target


def build_audit_prompt() -> str:  # noqa: D401
    """Return the system prompt used by audit mode.

    The assistant is instructed to inspect the local PBIP report project,
    evaluate it against a 90+ rule framework across seven domains, compute
    weighted scores, and write the final markdown report to
    ``AUDIT-REPORT.md`` in the current working directory.
    """
    return f"""\
You are a Power BI TMDL Audit Agent. Your task is to perform a comprehensive,
evidence-based audit of the local Power BI report project (PBIP format) in
the current working directory. Do NOT ask the user for any additional prompt
or clarification — work autonomously.

This audit targets the LOCAL report project only. Ignore anything specific to
Power BI Service online (workspace settings, deployment pipelines, gateway
configuration, dataflows, etc.).

===============================================================================
PROGRESS TRACKING AND RESUME SUPPORT
===============================================================================

A progress checklist file `{AUDIT_TODO_FILENAME}` exists in the current
working directory. It contains a checkbox task list that mirrors the audit
phases and every individual rule.

IMPORTANT — RESUME BEHAVIOUR:
If the audit was previously interrupted, `{AUDIT_TODO_FILENAME}` and
`{AUDIT_REPORT_FILENAME}` may already contain partial progress. At startup:
1. Read `{AUDIT_TODO_FILENAME}` and inspect which items are already `[x]`.
2. Read `{AUDIT_REPORT_FILENAME}` if it exists — it may contain sections
   already written by a previous run.
3. SKIP any phase/rule whose checkbox is already `[x]` — do NOT redo work.
4. Resume from the first unchecked `[ ]` item and continue normally.

INCREMENTAL PROGRESS RULES:
- After completing EACH phase or group of rules, update `{AUDIT_TODO_FILENAME}`
  using apply_patch to change `[ ]` to `[x]` for every completed item.
- Update the file incrementally — do NOT wait until the end.
- If a rule is not applicable, still mark it `[x]` (it was evaluated).
- You may add short notes below any checkbox line for context.

INCREMENTAL REPORT WRITING:
- You MUST also write findings to `{AUDIT_REPORT_FILENAME}` progressively,
  domain by domain, as you complete each phase.
- After Phase 1 (inventory), create `{AUDIT_REPORT_FILENAME}` with the
  report header and the "Audit Scope and Inventory" section.
- After each domain (Phases 2-8), APPEND that domain's "Findings by Domain"
  section to the report immediately using apply_patch.
- After all domains are done (Phase 9), add the final summary sections:
  Executive Summary, Score Card, Consolidated Findings Table, Priority
  Action Plan, and Confidence and Known Gaps.
- This way, if the agent is interrupted, partial findings are already saved
  in `{AUDIT_REPORT_FILENAME}` and can be continued.

This keeps progress visible and ensures no work is lost on long audits.

===============================================================================
PHASE 1 — DISCOVERY AND INVENTORY
===============================================================================

Scan the project folder recursively. Identify and catalogue:

- definition/database.tmdl ............... compatibility level
- definition/model.tmdl .................. model properties, culture, ref list
- definition/relationships.tmdl .......... all relationships
- definition/expressions.tmdl ............ shared M expressions / parameters
- definition/tables/*.tmdl ............... one file per table (columns, measures,
  partitions, hierarchies, calculated columns, calculation groups)
- definition/roles/*.tmdl ................ RLS / OLS roles
- definition/perspectives/*.tmdl ......... perspectives
- definition/cultures/*.tmdl ............. translations
- *.PBI  ................................. report and project manifests
- Report theme JSON files, images, custom visuals

Summarize what is present and what is missing.

===============================================================================
PHASE 2 — TMDL FORMAT REFERENCE (for your parsing)
===============================================================================

TMDL uses indentation-based syntax:
- Properties use `key: value`.
- Expressions use `key = <multiline indented block>`.
- `///` triple-slash above an object is its description.
- Boolean shorthand: writing `isHidden` alone means true.
- Object names with special characters are in single quotes.
- Calculated columns have a DAX `=` expression instead of `sourceColumn`.
- Relationships live in relationships.tmdl with fromColumn, toColumn,
  crossFilteringBehavior, securityFilteringBehavior, fromCardinality,
  toCardinality, isActive.

===============================================================================
PHASE 3 — SEVEN-DOMAIN AUDIT
===============================================================================

Evaluate every applicable rule below. For each finding, record:
  Rule ID, Severity (1-5), Domain, Object(s) affected, Evidence (quote from
  file), Impact description, and Recommendation.

Skip rules that are not applicable to the model (e.g., no RLS rules if no
roles exist). State "Not applicable" briefly.

-----------------------------------------------------------------------
DOMAIN 1: STRUCTURE AND STAR SCHEMA (weight 1.5)
-----------------------------------------------------------------------
STR-01  [Sev 3] No relationships defined (single-table or disconnected model).
STR-02  [Sev 3] Snowflake schema detected (dimension-to-dimension chains where
        a one-side table has a many-side relationship to another dimension).
STR-03  [Sev 3] Tables with no relationships (except intentional disconnected
        or parameter tables).
STR-04  [Sev 2] Wide table — >30 columns (warning); >60 columns (critical,
        elevate to Sev 3).
STR-05  [Sev 2] Fact-table columns visible to report users (all fact-table
        columns should be hidden, only measures exposed).

-----------------------------------------------------------------------
DOMAIN 2: MODELING — RELATIONSHIPS AND NAMING (weight 1.5)
-----------------------------------------------------------------------
MOD-01  [Sev 3] Bidirectional cross-filtering (`crossFilteringBehavior:
        bothDirections`). Exception: one-to-one, bridge tables, RLS propagation.
MOD-02  [Sev 3] Many-to-many relationships (`fromCardinality: many` +
        `toCardinality: many`).
MOD-03  [Sev 1] Inactive relationships (`isActive: false`) — review intent.
MOD-04  [Sev 2] Text-type join columns (`dataType: string` instead of int64).
MOD-05  [Sev 2] Bidirectional security filtering without any RLS role defined.
MOD-06  [Sev 1] Object names not capitalized (tables, columns, measures).
MOD-07  [Sev 1] Table names with "Dim" or "Fact" prefixes.
MOD-08  [Sev 2] Leading / trailing whitespace in object names.
MOD-09  [Sev 2] Special characters (tabs, line feeds) in names.
MOD-10  [Sev 2] Column references unqualified in DAX (should be
        'Table'[Column]).
MOD-11  [Sev 2] Measure references qualified in DAX (should be [Measure] not
        'Table'[Measure]).
MOD-12  [Sev 2] Tables with >10 visible items and no display folders.
MOD-13  [Sev 1] Hidden columns not using CamelCase convention.

-----------------------------------------------------------------------
DOMAIN 3: PERFORMANCE (weight 1.5)
-----------------------------------------------------------------------
Thresholds (Good / Warning / Critical):
  Columns per table:         <20 / 20-40 / >40
  Total columns in model:    <100 / 100-300 / >300
  Total tables:              <15 / 15-30 / >30
  Total measures:            <100 / 100-300 / >300
  Bidirectional rels:        0-1 / 2-3 / >3
  Many-to-many rels:         0 / 1-2 / >2
  Calculated columns total:  0-5 / 5-15 / >15
  Calculated tables total:   0-2 / 2-5 / >5

PRF-01  [Sev 2] Calculated column using RELATED — move to Power Query merge.
PRF-02  [Sev 3] Calculated column with aggregation function (SUM, CALCULATE,
        AVERAGE, COUNT, COUNTROWS) — should be a measure.
PRF-03  [Sev 2] Calculated tables (except date tables and field parameters).
PRF-04  [Sev 3] Auto date/time tables present (causes model bloat).
PRF-05  [Sev 3] Date table not marked with `DataCategory: Time`.
PRF-06  [Sev 3] DateTime columns with non-midnight time values — split into
        separate Date and Time columns.
PRF-07  [Sev 1] `isAvailableInMdx: false` not set on non-attribute columns.
PRF-08  [Sev 2] Numeric columns without `summarizeBy: none`.
PRF-09  [Sev 3] Floating-point `dataType: double` — use `decimal` instead.
PRF-10  [Sev 2] DirectQuery model without aggregation tables.

-----------------------------------------------------------------------
DOMAIN 4: SECURITY (weight 3.0)
-----------------------------------------------------------------------
SEC-01  [Sev 5] Hardcoded credentials in M expressions or connection strings
        (Password=, pwd=, key=).
SEC-02  [Sev 3] Local file paths (C:\\, UNC paths) in partition sources.
SEC-03  [Sev 3] Hardcoded server/database names instead of parameterized
        expressions.
SEC-04  [Sev 3] PII column name patterns (SSN, Email, CreditCard, BirthDate)
        without RLS defined.
SEC-05  [Sev 3] USERNAME() in RLS instead of USERPRINCIPALNAME().
SEC-06  [Sev 3] RLS filters applied directly on fact tables (many-side).
SEC-07  [Sev 2] LOOKUPVALUE in RLS expressions.
SEC-08  [Sev 2] Empty roles (modelPermission: read but no tablePermission).
SEC-09  [Sev 1] Hardcoded filter values in RLS (static values instead of
        security table lookup).
SEC-10  [Sev 2] Overly complex RLS expressions (>200 chars or >5 function
        calls).
SEC-11  [Sev 3] Measures referencing OLS-hidden columns (become unavailable
        to restricted users).
SEC-12  [Sev 3] OLS combined with RLS on the same table (conflict risk).

-----------------------------------------------------------------------
DOMAIN 5: DAX QUALITY (weight 1.5)
-----------------------------------------------------------------------
DAX-01  [Sev 3] FILTER on entire table inside CALCULATE — filter columns, not
        tables. Pattern: FILTER\\s*\\(\\s*'[^']+'\\s*,
DAX-02  [Sev 2] Division operator `/` between columns/measures instead of
        DIVIDE(). Pattern: ]\\s*/\\s*\\[
DAX-03  [Sev 2] FORMAT() in measures — returns text, kills VertiPaq
        optimization. Use formatString property.
DAX-04  [Sev 2] IFERROR / ISERROR usage — forces full evaluation.
DAX-05  [Sev 2] EARLIER / EARLIEST usage — replace with VAR/RETURN.
DAX-06  [Sev 2] Nested IF >= 4 deep — refactor to SWITCH(TRUE(), ...).
DAX-07  [Sev 2] ALL() on entire table without column specification.
DAX-08  [Sev 1] COUNTROWS(VALUES()) instead of DISTINCTCOUNT.
DAX-09  [Sev 1] HASONEVALUE + VALUES instead of SELECTEDVALUE.
DAX-10  [Sev 1] Iterator on single column instead of aggregator (e.g.,
        SUMX('T', 'T'[Col]) instead of SUM('T'[Col])).
DAX-11  [Sev 3] Nested iterators (e.g., SUMX(..., SUMX(...))).
DAX-12  [Sev 2] Measures >200 chars with no VAR and repeated sub-expressions.
DAX-13  [Sev 2] Measure complexity: >30 lines, >1000 chars, or >10 distinct
        DAX functions.
DAX-14  [Sev 2] Nested CALCULATE (CALCULATE inside CALCULATE).
DAX-15  [Sev 2] Measure dependency chain deeper than 5 levels (Sev 3 if >10).

-----------------------------------------------------------------------
DOMAIN 6: METADATA AND DOCUMENTATION (weight 1.0)
-----------------------------------------------------------------------
MET-01  [Sev 2] Measures without descriptions.
MET-02  [Sev 1] Visible columns without descriptions.
MET-03  [Sev 1] Tables without descriptions.
MET-04  [Sev 1] Description duplicates the object name.
MET-05  [Sev 1] Description shorter than 10 characters.
MET-06  [Sev 2] Currency measures (name contains Sales, Revenue, Amount, Cost,
        Price) without currency format string.
MET-07  [Sev 2] Percentage measures (name contains Ratio, Percent, Rate, %)
        without percentage format string.
MET-08  [Sev 2] Date columns without explicit format string.
MET-09  [Sev 2] Visible numeric measures without any formatString.
MET-10  [Sev 1] `PBI_FormatHint` annotation with `isGeneralNumber: true`
        (no explicit format set).

-----------------------------------------------------------------------
DOMAIN 7: ANTI-PATTERNS AND HIDDEN FIELDS (weight 1.0)
-----------------------------------------------------------------------
APT-01  [Sev 2] Foreign key columns not hidden (many-side of relationships).
APT-02  [Sev 2] Key/ID columns without `summarizeBy: none`.
APT-03  [Sev 2] Hidden columns not referenced in any relationship, hierarchy,
        sortByColumn, or DAX — candidates for removal.
APT-03b [Sev 2] Unused imported columns — any column (visible or hidden) with
        a `sourceColumn` that is NOT referenced in: any relationship
        (fromColumn/toColumn), any hierarchy level, any sortByColumn, any
        measure or calculated-column DAX expression, any RLS/OLS
        tablePermission expression, or any report visual field. These columns
        inflate model size without adding value. Recommend removing from the
        Power Query output or hiding and marking for deprecation.
APT-04  [Sev 3] Missing lineageTags on objects.
APT-05  [Sev 4] Duplicate lineageTags across objects (copy-paste error).
APT-06  [Sev 3] Invalid lineageTag GUID format.

===============================================================================
PHASE 4 — SCORING
===============================================================================

Use the BPA 1-5 severity scale:
  1 = Cosmetic (informational)
  2 = Minor (warning — may cause end-user confusion)
  3 = Important (error — performance degradation or functional issues)
  4 = Very Important (error — high risk)
  5 = Critical (error — guaranteed deployment or logical errors)

Apply category weights:
  Security ............. x3.0
  Performance .......... x1.5
  DAX Quality .......... x1.5
  Modeling ............. x1.5
  Structure ............ x1.5
  Documentation ........ x1.0
  Anti-patterns ........ x1.0

Compute weighted score:
  Score = (Sum of passed_rules x category_weight)
        / (Sum of total_applicable_rules x category_weight) x 100

Letter grades:
  A (90-100%) — Production-ready
  B (80-89%)  — Minor improvements needed
  C (70-79%)  — Several issues to address
  D (60-69%)  — Significant improvements needed
  F (<60%)    — Critical issues present

===============================================================================
PHASE 5 — OUTPUT REQUIREMENTS
===============================================================================

`{AUDIT_REPORT_FILENAME}` is written INCREMENTALLY as you progress through
the audit. Use apply_patch to create and append content step by step:

STEP A — After Phase 1 (Inventory), create the file with:

  # Power BI TMDL Audit Report

  ## 2. Audit Scope and Inventory
  - Files discovered (table count, measure count, relationship count, roles, etc.)
  - Model storage mode (Import / DirectQuery / Composite)
  - Compatibility level
  - What was audited vs. what was not found

STEP B — After EACH domain (Phases 2-8), append the domain findings section:

  ## 4. Findings by Domain

  ### Domain N: <Domain Name>

  **Rule-ID** — <Short title> (Severity N)
  - **Object(s):** <table/column/measure affected>
  - **Evidence:** <quote or reference from the TMDL file>
  - **Impact:** <what goes wrong>
  - **Recommendation:** <specific fix>

  If a domain has zero findings, state: "No issues detected."

STEP C — After all domains are done (Phase 9), prepend / append the summary
sections to complete the report:

  ## 1. Executive Summary
  One paragraph: overall grade (letter + percentage), model name, top 3 risks,
  and whether the model is production-ready.
  (Insert this at the TOP of the report, right after the title.)

  ## 3. Score Card
  A summary table with one row per domain:

  | Domain | Applicable Rules | Passed | Failed | Weight | Weighted Score |
  |--------|-----------------|--------|--------|--------|----------------|
  | Structure | ... | ... | ... | x1.5 | ... |
  | Modeling | ... | ... | ... | x1.5 | ... |
  | Performance | ... | ... | ... | x1.5 | ... |
  | Security | ... | ... | ... | x3.0 | ... |
  | DAX Quality | ... | ... | ... | x1.5 | ... |
  | Metadata | ... | ... | ... | x1.0 | ... |
  | Anti-Patterns | ... | ... | ... | x1.0 | ... |
  | **Overall** | **...** | **...** | **...** | | **Grade: X (nn%)** |

  ## 5. Consolidated Findings Table
  A single table of ALL findings across all domains, sorted by severity
  (highest first):

  | # | Rule ID | Severity | Domain | Object(s) | Finding | Recommendation |
  |---|---------|----------|--------|-----------|---------|----------------|
  | 1 | SEC-01 | 5 | Security | ... | ... | ... |
  | 2 | ... | ... | ... | ... | ... | ... |

  ## 6. Priority Action Plan
  Group recommendations into three tiers:
  - **Now (Critical/High — Sev 3-5):** Must fix before production.
  - **Next (Medium — Sev 2):** Fix in next iteration.
  - **Later (Low — Sev 1):** Style and cosmetic improvements.

  ## 7. Confidence and Known Gaps
  - State which areas could not be fully evaluated and why.
  - Note any assumptions made.

===============================================================================
EXECUTION INSTRUCTIONS
===============================================================================

1. Read `{AUDIT_TODO_FILENAME}`. If items are already checked `[x]`, this is
   a RESUMED audit — skip completed phases and pick up from the first `[ ]`.
2. Read `{AUDIT_REPORT_FILENAME}` if it exists to see what was already written.
3. Read every project file methodically. Do NOT skip files.
4. After Phase 1 (inventory):
   - Create or update `{AUDIT_REPORT_FILENAME}` with the Inventory section.
   - Check off Phase 1 items in `{AUDIT_TODO_FILENAME}`.
5. For each domain (Phases 2-8):
   - Evaluate every rule in the domain.
   - APPEND that domain's findings to `{AUDIT_REPORT_FILENAME}` immediately.
   - Check off the completed rules in `{AUDIT_TODO_FILENAME}`.
   - Move to the next domain only after both files are updated.
6. Ground EVERY finding in observed file evidence. If evidence is missing,
   state "Not found" — do NOT invent findings.
7. After all domains (Phase 9):
   - Compute the score.
   - Add Executive Summary, Score Card, Consolidated Findings Table,
     Priority Action Plan, and Confidence sections to `{AUDIT_REPORT_FILENAME}`.
   - Check off Phase 9 items in `{AUDIT_TODO_FILENAME}`.
8. After writing the file, respond briefly with confirmation, the overall
   grade, and the top 3 risks."""



"""Custom web server that extends textual-serve with favicon support.

This module subclasses ``textual_serve.server.Server`` to:

1. Use a project-local HTML template (with a ``<link rel="icon">`` tag).
2. Serve a custom favicon at ``/favicon.ico``.

It is invoked as ``python -m pbi_agent.web.serve <command> [options]`` by
:func:`pbi_agent.cli._handle_web_command`.
"""

from __future__ import annotations

import argparse
from pathlib import Path

from aiohttp import web
from textual_serve.server import Server

from pbi_agent.branding import startup_panel

_WEB_DIR = Path(__file__).resolve().parent
_TEMPLATES_DIR = str(_WEB_DIR / "templates")
_FAVICON_PATH = _WEB_DIR / "static" / "favicon.png"


class _FaviconServer(Server):
    """A ``textual-serve`` server with custom branding, template, and favicon."""

    def __init__(
        self,
        command: str,
        host: str = "localhost",
        port: int = 8000,
        title: str | None = None,
        public_url: str | None = None,
    ) -> None:
        super().__init__(
            command,
            host=host,
            port=port,
            title=title,
            public_url=public_url,
            templates_path=_TEMPLATES_DIR,
        )

    # ------------------------------------------------------------------
    async def _make_app(self) -> web.Application:  # type: ignore[override]
        app = await super()._make_app()
        app.router.add_get("/favicon.ico", self._handle_favicon)
        return app

    async def on_startup(self, app: web.Application) -> None:
        del app
        self.console.print(startup_panel(), highlight=False)
        self.console.print(f"  Serving on [bold]{self.public_url}[/bold]")
        self.console.print("[cyan]  Press Ctrl+C to quit[/cyan]")

    async def _handle_favicon(self, _request: web.Request) -> web.FileResponse:
        return web.FileResponse(
            _FAVICON_PATH,
            headers={"Content-Type": "image/png"},
        )


# -- CLI entry-point ----------------------------------------------------


def _parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="pbi-agent web server")
    parser.add_argument("command", help="Shell command to run the Textual app")
    parser.add_argument("--host", default="localhost")
    parser.add_argument("--port", type=int, default=8000)
    parser.add_argument("--title", default=None)
    parser.add_argument("--url", default=None, dest="public_url")
    parser.add_argument("--dev", action="store_true", default=False)
    return parser.parse_args(argv)


def main(argv: list[str] | None = None) -> None:
    args = _parse_args(argv)
    server = _FaviconServer(
        command=args.command,
        host=args.host,
        port=args.port,
        title=args.title,
        public_url=args.public_url,
    )
    server.serve(debug=args.dev)


if __name__ == "__main__":
    main()



"""Anthropic Messages API provider.

Uses direct HTTP calls (``urllib.request``) to the Anthropic Messages API.
Conversation history is managed client-side by maintaining a full
``messages`` list that is sent with every request.

All tools (including shell and apply_patch) are registered function tools
— no provider-specific native tool types.
"""

from __future__ import annotations
import json
import logging
import time
import urllib.error
import urllib.request
from typing import Any

from pbi_agent.agent.system_prompt import get_system_prompt
from pbi_agent.agent.tool_runtime import execute_tool_calls as _execute_tool_calls
from pbi_agent.config import Settings
from pbi_agent.models.messages import (
    CompletedResponse,
    ImageAttachment,
    TokenUsage,
    ToolCall,
    UserTurnInput,
    WebSearchSource,
)
from pbi_agent.providers.base import Provider
from pbi_agent.session_store import MessageRecord
from pbi_agent.tools.registry import get_anthropic_tool_definitions
from pbi_agent.tools.types import ToolContext
from pbi_agent.ui.display_protocol import DisplayProtocol

_log = logging.getLogger(__name__)

ANTHROPIC_API_URL = "https://api.anthropic.com/v1/messages"
ANTHROPIC_VERSION = "2023-06-01"

# Map the CLI reasoning-effort values to Anthropic adaptive thinking effort.
_EFFORT_MAP: dict[str, str] = {
    "low": "low",
    "medium": "medium",
    "high": "high",
    "xhigh": "max",
}

_HTTP_ERROR_TYPES: dict[int, str] = {
    400: "invalid_request_error",
    401: "authentication_error",
    403: "permission_error",
    404: "not_found_error",
    413: "request_too_large",
    429: "rate_limit_error",
    500: "api_error",
    529: "overloaded_error",
}

_HTTP_ERROR_MESSAGES: dict[int, str] = {
    400: "There was an issue with the format or content of your request.",
    401: "There's an issue with your API key.",
    403: "Your API key does not have permission to use the specified resource.",
    404: "The requested resource could not be found.",
    413: "Request exceeds the maximum allowed number of bytes.",
    429: "Your account has hit a rate limit.",
    500: "An unexpected error has occurred internal to Anthropic's systems.",
    529: "The API is temporarily overloaded.",
}


class AnthropicProvider(Provider):
    """Provider backed by the Anthropic Messages HTTP API."""

    def __init__(
        self,
        settings: Settings,
        *,
        system_prompt: str | None = None,
        excluded_tools: set[str] | None = None,
    ) -> None:
        self._settings = settings
        self._tools = get_anthropic_tool_definitions(excluded_names=excluded_tools)
        if settings.web_search:
            self._tools.append(_anthropic_web_search_tool(settings.model))
        self._system_prompt = system_prompt or get_system_prompt()
        # Client-side conversation history — full messages list.
        self._messages: list[dict[str, Any]] = []

    @property
    def settings(self) -> Settings:
        return self._settings

    # -- lifecycle -----------------------------------------------------------

    def connect(self) -> None:
        # HTTP is stateless; nothing to connect.  We validate the API key
        # eagerly so errors surface early.
        if not self._settings.api_key:
            raise ValueError(
                "Missing API key. Set PBI_AGENT_API_KEY in environment or pass "
                "--api-key."
            )

    def close(self) -> None:
        pass

    def reset_conversation(self) -> None:
        self._messages.clear()

    def restore_messages(self, messages: list[MessageRecord]) -> None:
        self._messages = [
            {
                "role": message.role,
                "content": [{"type": "text", "text": message.content}],
            }
            for message in messages
            if message.role in {"user", "assistant"} and message.content
        ]

    # -- request_turn --------------------------------------------------------

    def request_turn(
        self,
        *,
        user_message: str | None = None,
        user_input: UserTurnInput | None = None,
        tool_result_items: list[dict[str, Any]] | None = None,
        instructions: str | None = None,
        display: DisplayProtocol,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
    ) -> CompletedResponse:
        # Build the new message to append to history.
        if user_input is None and user_message is not None:
            user_input = UserTurnInput(text=user_message)

        if user_input is not None:
            self._messages.append(
                {
                    "role": "user",
                    "content": _anthropic_user_content_blocks(user_input),
                }
            )
        elif tool_result_items is not None:
            # Tool results are sent as a user message containing tool_result
            # content blocks.
            self._messages.append(
                {
                    "role": "user",
                    "content": tool_result_items,
                }
            )
        else:
            raise ValueError("Either user_input or tool_result_items is required")

        system_prompt = instructions or self._system_prompt

        response = self._http_request(
            system_prompt=system_prompt,
            display=display,
        )
        session_usage.add(response.usage)
        turn_usage.add(response.usage)
        display.session_usage(session_usage)
        return response

    # -- execute_tool_calls --------------------------------------------------

    def execute_tool_calls(
        self,
        response: CompletedResponse,
        *,
        max_workers: int,
        display: DisplayProtocol,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
        sub_agent_depth: int = 0,
    ) -> tuple[list[dict[str, Any]], bool]:
        """Execute all tool calls in the response.

        All ``tool_use`` content blocks are routed through the registered
        function tool handlers via ``tool_runtime.execute_tool_calls``.
        """
        # The raw content blocks are stored in provider_data.
        pdata = response.provider_data or {}
        content_blocks: list[dict[str, Any]] = (
            pdata.get("content_blocks", []) if isinstance(pdata, dict) else pdata
        )
        tool_use_blocks = [b for b in content_blocks if b.get("type") == "tool_use"]

        if not tool_use_blocks:
            return [], False

        # Convert Anthropic tool_use blocks to ToolCall objects so we can
        # reuse the existing tool_runtime.
        fn_calls = [
            ToolCall(
                call_id=b.get("id", ""),
                name=b.get("name", ""),
                arguments=b.get("input"),
            )
            for b in tool_use_blocks
        ]

        displayable_calls = [call for call in fn_calls if call.name != "sub_agent"]
        if displayable_calls:
            display.function_start(len(displayable_calls))
        batch = _execute_tool_calls(
            fn_calls,
            max_workers=max_workers,
            context=ToolContext(
                settings=self._settings,
                display=display,
                session_usage=session_usage,
                turn_usage=turn_usage,
                sub_agent_depth=sub_agent_depth,
            ),
        )
        had_errors = batch.had_errors

        tool_result_items: list[dict[str, Any]] = []
        for result in batch.results:
            call = _find_by_id(fn_calls, result.call_id)
            if not (call and call.name == "sub_agent"):
                display.function_result(
                    name=call.name if call else "unknown",
                    success=not result.is_error,
                    call_id=result.call_id,
                    arguments=call.arguments if call else None,
                )
            tool_result_items.append(
                {
                    "type": "tool_result",
                    "tool_use_id": result.call_id,
                    "content": _anthropic_tool_result_content(result),
                    **({"is_error": True} if result.is_error else {}),
                }
            )
        if displayable_calls:
            display.tool_group_end()

        return tool_result_items, had_errors

    # -- HTTP transport ------------------------------------------------------

    def _http_request(
        self,
        *,
        system_prompt: str | None,
        display: DisplayProtocol,
    ) -> CompletedResponse:
        """Send the current messages to the Anthropic Messages API and return
        a parsed ``CompletedResponse``."""
        display.wait_start("waiting for Anthropic response...")

        body: dict[str, Any] = {
            "model": self._settings.model,
            "max_tokens": self._settings.max_tokens,
            "cache_control": {"type": "ephemeral"},
            "tools": self._tools,
            "messages": self._messages,
        }

        if _supports_adaptive_thinking(self._settings.model):
            body["thinking"] = {"type": "adaptive"}
            effort = _EFFORT_MAP.get(self._settings.reasoning_effort, "high")
            body["output_config"] = {"effort": effort}

        if system_prompt:
            body["system"] = system_prompt

        request_data = json.dumps(body).encode("utf-8")

        headers = {
            "Content-Type": "application/json",
            "x-api-key": self._settings.api_key,
            "anthropic-version": ANTHROPIC_VERSION,
        }

        max_retries = self._settings.max_retries
        last_error: Exception | None = None
        last_error_message: str | None = None

        for attempt in range(max_retries + 1):
            if attempt > 0:
                display.retry_notice(attempt, max_retries)

            try:
                req = urllib.request.Request(
                    ANTHROPIC_API_URL,
                    data=request_data,
                    headers=headers,
                    method="POST",
                )
                with urllib.request.urlopen(req, timeout=300) as resp:
                    response_bytes = resp.read()
                    response_json = json.loads(response_bytes.decode("utf-8"))

                result = self._parse_response(response_json)
                display.wait_stop()

                # Render thinking blocks (if any) before the main text.
                pdata = result.provider_data or {}
                if isinstance(pdata, dict):
                    for thinking_text in pdata.get("thinking_parts", []):
                        display.render_thinking(thinking_text)
                    if pdata.get("has_redacted_thinking"):
                        display.render_redacted_thinking()

                display_items = pdata.get("display_items", [])
                if isinstance(display_items, list) and display_items:
                    for item in display_items:
                        if not isinstance(item, dict):
                            continue
                        item_type = item.get("type")
                        if item_type == "text":
                            text = item.get("text")
                            if isinstance(text, str) and text:
                                display.render_markdown(text)
                        elif item_type == "web_search":
                            _display_web_search_result(
                                display,
                                item.get("sources", []),
                                queries=item.get("queries", []),
                            )
                else:
                    if result.text:
                        display.render_markdown(result.text)

                    if result.web_search_sources:
                        display.web_search_sources(result.web_search_sources)

                # Append the assistant's response to conversation history so
                # subsequent turns include it.
                self._messages.append(
                    {
                        "role": "assistant",
                        "content": response_json.get("content", []),
                    }
                )

                return result

            except urllib.error.HTTPError as exc:
                error_body = _read_error_body(exc)
                error_payload = _normalize_http_error(exc, error_body)

                # Rate limiting
                if exc.code == 429:
                    if attempt >= max_retries:
                        display.wait_stop()
                        raise RuntimeError(
                            _format_error_message(
                                f"Anthropic rate limit exceeded after {max_retries + 1} attempts",
                                error_payload,
                            )
                        ) from exc

                    wait = _extract_retry_after(exc, attempt)
                    display.rate_limit_notice(
                        wait_seconds=wait,
                        attempt=attempt + 1,
                        max_retries=max_retries,
                    )
                    time.sleep(wait)
                    continue

                # Overloaded (529)
                if exc.code == 529:
                    if attempt >= max_retries:
                        display.wait_stop()
                        raise RuntimeError(
                            _format_error_message(
                                f"Anthropic API overloaded after {max_retries + 1} attempts",
                                error_payload,
                            )
                        ) from exc
                    wait = min(2.0 * (2**attempt), 30.0) + 1.0
                    display.overload_notice(
                        wait_seconds=wait,
                        attempt=attempt + 1,
                        max_retries=max_retries,
                    )
                    time.sleep(wait)
                    continue

                # Server errors (5xx) — retry
                if exc.code >= 500:
                    last_error = exc
                    last_error_message = _format_error_message(
                        f"Anthropic request failed after {max_retries + 1} attempts",
                        error_payload,
                    )
                    continue

                # Client errors (4xx) — don't retry
                display.wait_stop()
                raise RuntimeError(
                    _format_error_message(
                        f"Anthropic API error {exc.code}",
                        error_payload,
                    )
                ) from exc

            except urllib.error.URLError as exc:
                last_error = exc
                last_error_message = None
                continue

        display.wait_stop()
        if last_error is not None:
            if last_error_message:
                raise RuntimeError(last_error_message) from last_error
            raise RuntimeError(
                f"Anthropic request failed after {max_retries + 1} attempts: "
                f"{last_error}"
            ) from last_error
        raise RuntimeError("Anthropic request failed after retries.")

    # -- response parsing ----------------------------------------------------

    def _parse_response(self, response_json: dict[str, Any]) -> CompletedResponse:
        """Parse an Anthropic Messages API response into a CompletedResponse."""
        content_blocks: list[dict[str, Any]] = response_json.get("content", [])

        text_parts: list[str] = []
        thinking_parts: list[str] = []
        has_redacted_thinking: bool = False
        function_calls: list[ToolCall] = []
        web_search_sources: list[WebSearchSource] = []
        display_items: list[dict[str, Any]] = []
        pending_web_search_queries: dict[str, list[str]] = {}

        for block in content_blocks:
            block_type = block.get("type")

            if block_type == "thinking":
                thinking_text = block.get("thinking", "")
                if thinking_text:
                    thinking_parts.append(thinking_text)

            elif block_type == "redacted_thinking":
                has_redacted_thinking = True

            elif block_type == "text":
                text = block.get("text", "")
                if text:
                    text_parts.append(text)
                    display_items.append({"type": "text", "text": text})

            elif block_type == "tool_use":
                name = block.get("name", "")
                function_calls.append(
                    ToolCall(
                        call_id=block.get("id", ""),
                        name=name,
                        arguments=block.get("input"),
                    )
                )

            elif block_type == "server_tool_use":
                if str(block.get("name", "")).startswith("web_search"):
                    pending_web_search_queries[block.get("id", "")] = (
                        _extract_anthropic_web_search_queries(block)
                    )

            elif block_type == "web_search_tool_result":
                sources_for_block: list[dict[str, str]] = []
                for entry in block.get("content", []):
                    if not isinstance(entry, dict):
                        continue
                    if entry.get("type") == "web_search_result":
                        source = WebSearchSource(
                            title=str(entry.get("title", "")),
                            url=str(entry.get("url", "")),
                            snippet=str(entry.get("page_snippet", "")),
                        )
                        web_search_sources.append(source)
                        sources_for_block.append(
                            {
                                "title": source.title,
                                "url": source.url,
                                "snippet": source.snippet,
                            }
                        )
                display_items.append(
                    {
                        "type": "web_search",
                        "queries": pending_web_search_queries.get(
                            str(block.get("tool_use_id", "")),
                            [],
                        ),
                        "sources": sources_for_block,
                    }
                )

        # Parse usage
        usage_obj = response_json.get("usage", {})
        base_input_tokens = int(usage_obj.get("input_tokens", 0) or 0)
        output_tokens = int(usage_obj.get("output_tokens", 0) or 0)
        cache_read_tokens = int(usage_obj.get("cache_read_input_tokens", 0) or 0)
        cache_creation_tokens = int(
            usage_obj.get("cache_creation_input_tokens", 0) or 0
        )

        # Break down cache creation by TTL (default is 5-minute).
        cache_creation_obj = usage_obj.get("cache_creation", {})
        if isinstance(cache_creation_obj, dict):
            cache_1h = int(cache_creation_obj.get("ephemeral_1h_input_tokens", 0) or 0)
        else:
            cache_1h = 0
        cache_5m = max(cache_creation_tokens - cache_1h, 0)

        # Total input for display (base + cache reads + cache writes).
        total_input = base_input_tokens + cache_read_tokens + cache_creation_tokens

        return CompletedResponse(
            response_id=response_json.get("id"),
            text="\n\n".join(text_parts).strip(),
            usage=TokenUsage(
                input_tokens=total_input,
                cached_input_tokens=cache_read_tokens,
                cache_write_tokens=cache_5m,
                cache_write_1h_tokens=cache_1h,
                output_tokens=output_tokens,
                context_tokens=total_input + output_tokens,
            ),
            function_calls=function_calls,
            # Store raw content blocks so execute_tool_calls can access the
            # full tool_use data (including input parameters), plus parsed
            # thinking data for display.
            provider_data={
                "content_blocks": content_blocks,
                "thinking_parts": thinking_parts,
                "has_redacted_thinking": has_redacted_thinking,
                "display_items": display_items,
            },
            web_search_sources=web_search_sources,
        )


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _find_by_id(calls: list[ToolCall], call_id: str) -> ToolCall | None:
    for c in calls:
        if c.call_id == call_id:
            return c
    return None


def _anthropic_user_content_blocks(user_input: UserTurnInput) -> list[dict[str, Any]]:
    blocks: list[dict[str, Any]] = []
    for image in user_input.images:
        blocks.append(_anthropic_image_block(image))
    if user_input.text:
        blocks.append({"type": "text", "text": user_input.text})
    return blocks or [{"type": "text", "text": ""}]


def _anthropic_tool_result_content(result) -> str | list[dict[str, Any]]:
    if not result.attachments:
        return result.output_json

    blocks = [_anthropic_image_block(image) for image in result.attachments]
    blocks.append({"type": "text", "text": result.output_json})
    return blocks


def _anthropic_image_block(image: ImageAttachment) -> dict[str, Any]:
    return {
        "type": "image",
        "source": {
            "type": "base64",
            "media_type": image.mime_type,
            "data": image.data_base64,
        },
    }


def _extract_anthropic_web_search_queries(block: dict[str, Any]) -> list[str]:
    raw_input = block.get("input")
    if not isinstance(raw_input, dict):
        return []
    raw_query = raw_input.get("query")
    if isinstance(raw_query, str) and raw_query.strip():
        return [raw_query.strip()]
    return []


def _supports_adaptive_thinking(model: str) -> bool:
    normalized = model.strip().lower()
    return not normalized.startswith("claude-haiku")


def _anthropic_web_search_tool(model: str) -> dict[str, Any]:
    tool: dict[str, Any] = {
        "type": "web_search_20260209",
        "name": "web_search",
    }
    if model.strip().lower().startswith("claude-haiku"):
        tool["allowed_callers"] = ["direct"]
    return tool


def _display_web_search_result(
    display: DisplayProtocol,
    sources: list[dict[str, Any]],
    *,
    queries: list[str] | None = None,
) -> None:
    display.function_start(1)
    display.function_result(
        name="web_search",
        success=True,
        call_id="",
        arguments={
            "queries": list(queries or []),
            "sources": list(sources),
        },
    )
    display.tool_group_end()


def _extract_retry_after(exc: urllib.error.HTTPError, attempt: int) -> float:
    """Extract retry wait from Retry-After header or use exponential backoff."""
    try:
        retry_header = exc.headers.get("Retry-After") if exc.headers else None
        if retry_header:
            return max(0.1, min(float(retry_header), 60.0)) + 1.0
    except (ValueError, TypeError):
        pass
    return min(2.0 * (2**attempt), 30.0) + 1.0


def _read_error_body(exc: urllib.error.HTTPError) -> str:
    try:
        return exc.read().decode("utf-8", errors="replace")
    except Exception:
        return ""


def _normalize_http_error(
    exc: urllib.error.HTTPError,
    error_body: str,
) -> dict[str, Any]:
    payload = _parse_error_payload(error_body)
    error_type = _HTTP_ERROR_TYPES.get(exc.code)
    message = _HTTP_ERROR_MESSAGES.get(exc.code, f"HTTP {exc.code}")
    request_id = _request_id_from_headers(exc)

    if payload is not None:
        payload_request_id = payload.get("request_id")
        if isinstance(payload_request_id, str) and payload_request_id.strip():
            request_id = payload_request_id.strip()

        error_value = payload.get("error")
        if isinstance(error_value, dict):
            payload_type = error_value.get("type")
            if isinstance(payload_type, str) and payload_type.strip():
                error_type = payload_type.strip()
            payload_message = error_value.get("message")
            if isinstance(payload_message, str) and payload_message.strip():
                message = payload_message.strip()
        elif isinstance(error_value, str) and error_value.strip():
            message = error_value.strip()

    if error_type is None:
        if 400 <= exc.code < 500:
            error_type = "invalid_request_error"
        else:
            error_type = "api_error"

    return {
        "type": "error",
        "status": exc.code,
        "error": {
            "type": error_type,
            "message": message,
        },
        **({"request_id": request_id} if request_id else {}),
    }


def _parse_error_payload(error_body: str) -> dict[str, Any] | None:
    stripped = error_body.strip()
    if not stripped.startswith("{"):
        return None
    try:
        payload = json.loads(stripped)
    except json.JSONDecodeError:
        return None
    return payload if isinstance(payload, dict) else None


def _request_id_from_headers(exc: urllib.error.HTTPError) -> str | None:
    if not exc.headers:
        return None
    request_id = exc.headers.get("request-id")
    if isinstance(request_id, str) and request_id.strip():
        return request_id.strip()
    return None


def _format_error_message(prefix: str, error_payload: dict[str, Any]) -> str:
    return f"{prefix}: {json.dumps(error_payload, sort_keys=True)}"
